/* Copyright (C) 2007-2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Igor Kulakov [committer], Maksym Zyzak, Sergei Zharko */


/// \file   CaTrackParam.h
/// \brief  header file for the ca::TrackParam class
/// \since  02.06.2022
/// \author Sergey Gorbunov

#ifndef CA_CORE_CaTrackParam_h
#define CA_CORE_CaTrackParam_h 1

#include "CaConstants.h"
#include "CaSimd.h"
#include "CaUtils.h"

#include <boost/serialization/access.hpp>

#include <string>

namespace cbm::algo::ca
{
  /// \class cbm::algo::ca:: TrackParamBase
  /// \brief Class represents track parameters in the CA tracking algorithm
  ///
  /// This is a template that can be instantiated for different floating point types.
  /// Track parameters: variable parameterts {x, y, tx, ty, q/p, t, vi} at fixed z
  /// Covariation matrix: low-diagopnal C[28]
  ///
  template<typename DataT>
  class TrackParamBase {

   public:
    friend class boost::serialization::access;

    static constexpr int kNtrackParam{7};  ///<  N of variable track parameters: {x, y, tx, ty, q/p, t, vi}
    static constexpr int kNcovParam{(kNtrackParam) * (kNtrackParam + 1) / 2};  ///< N of covariance matrix parameters

    typedef std::array<DataT, kNcovParam> CovMatrix_t;  ///< covariance matrix type

    TrackParamBase() = default;

    template<typename T>
    TrackParamBase(const TrackParamBase<T>& tr)
    {
      Set(tr);
    }

    /// Set all SIMD entries from all SIMD entries of the other class
    /// It works for scalar and fvec types,
    /// except of the case when DataT is scalar and TdataB is fvec.
    template<typename TdataB>
    void Set(const TrackParamBase<TdataB>& Tb)
    {
      CopyBase<TdataB, true, true>(0, Tb, 0);
    }

    /// Set all SIMD entries from one SIMD entry of the other class
    /// It also works when DataT is scalar
    void Set(const TrackParamBase<fvec>& Tb, const int ib) { CopyBase<fvec, true, false>(0, Tb, ib); }

    /// Set one SIMD entry from one SIMD entry of the other class
    /// It only works when DataT is fvec, TdataB is scalar
    template<typename TdataB>
    void SetOneEntry(const int ia, const TrackParamBase<TdataB>& Tb)
    {
      CopyBase<TdataB, false, true>(ia, Tb, 0);
    }

    /// Set one SIMD entry from one SIMD entry of the other class
    /// It only works when DataT is fvec, TdataB is fvec
    void SetOneEntry(const int ia, const TrackParamBase<fvec>& Tb, const int ib)
    {
      CopyBase<fvec, false, false>(ia, Tb, ib);
    }

    /// ---------------------------------------------------------------------------------------------------------------------
    /// Getters without 'Get' prefix to ease the math formulae
    ///

    /// \brief Gets z position [cm]
    DataT Z() const { return fZ; }

    /// \brief Gets x position [cm]
    DataT X() const { return fX; }

    /// \brief Gets y position [cm]
    DataT Y() const { return fY; }

    /// \brief Gets slope along x-axis
    DataT Tx() const { return fTx; }

    /// \brief Gets slope along y-axis
    DataT Ty() const { return fTy; }

    /// \brief Gets charge over momentum [ec/GeV]
    DataT Qp() const { return fQp; }

    /// \brief Gets time [ns]
    DataT Time() const { return fT; }

    /// \brief Gets inverse velocity [ns/cm]
    DataT Vi() const { return fVi; }

    /// \brief Gets Chi-square of track fit model
    DataT ChiSq() const { return fChiSq; }

    /// \brief Gets NDF of track fit model
    DataT Ndf() const { return fNdf; }

    /// \brief Gets Chi-square of time measurements
    DataT ChiSqTime() const { return fChiSqTime; }

    /// \brief Gets NDF of time measurements
    DataT NdfTime() const { return fNdfTime; }

    /// \brief Get covariance matrix element
    /// \param i row
    /// \param j column
    /// \return covariance matrix element
    DataT C(int i, int j) const
    {
      int ind = (j <= i) ? i * (1 + i) / 2 + j : j * (1 + j) / 2 + i;
      return fCovMatrix[ind];
    }

    /// \brief Get covariance matrix element when indices are known at compile time
    /// \param i row
    /// \param j column
    /// \return matrix element
    template<int i, int j>
    DataT C() const
    {
      constexpr int ind = (j <= i) ? i * (1 + i) / 2 + j : j * (1 + j) / 2 + i;
      return fCovMatrix[ind];
    }

    /// \brief Individual getters for covariance matrix elements
    /// \return covariance matrix element
    ///
    DataT C00() const { return C<0, 0>(); }
    DataT C01() const { return C<0, 1>(); }
    DataT C02() const { return C<0, 2>(); }
    DataT C03() const { return C<0, 3>(); }
    DataT C04() const { return C<0, 4>(); }
    DataT C05() const { return C<0, 5>(); }
    DataT C06() const { return C<0, 6>(); }

    DataT C10() const { return C<1, 0>(); }
    DataT C11() const { return C<1, 1>(); }
    DataT C12() const { return C<1, 2>(); }
    DataT C13() const { return C<1, 3>(); }
    DataT C14() const { return C<1, 4>(); }
    DataT C15() const { return C<1, 5>(); }
    DataT C16() const { return C<1, 6>(); }

    DataT C20() const { return C<2, 0>(); }
    DataT C21() const { return C<2, 1>(); }
    DataT C22() const { return C<2, 2>(); }
    DataT C23() const { return C<2, 3>(); }
    DataT C24() const { return C<2, 4>(); }
    DataT C25() const { return C<2, 5>(); }
    DataT C26() const { return C<2, 6>(); }

    DataT C30() const { return C<3, 0>(); }
    DataT C31() const { return C<3, 1>(); }
    DataT C32() const { return C<3, 2>(); }
    DataT C33() const { return C<3, 3>(); }
    DataT C34() const { return C<3, 4>(); }
    DataT C35() const { return C<3, 5>(); }
    DataT C36() const { return C<3, 6>(); }

    DataT C40() const { return C<4, 0>(); }
    DataT C41() const { return C<4, 1>(); }
    DataT C42() const { return C<4, 2>(); }
    DataT C43() const { return C<4, 3>(); }
    DataT C44() const { return C<4, 4>(); }
    DataT C45() const { return C<4, 5>(); }
    DataT C46() const { return C<4, 6>(); }

    DataT C50() const { return C<5, 0>(); }
    DataT C51() const { return C<5, 1>(); }
    DataT C52() const { return C<5, 2>(); }
    DataT C53() const { return C<5, 3>(); }
    DataT C54() const { return C<5, 4>(); }
    DataT C55() const { return C<5, 5>(); }
    DataT C56() const { return C<5, 6>(); }

    DataT C60() const { return C<6, 0>(); }
    DataT C61() const { return C<6, 1>(); }
    DataT C62() const { return C<6, 2>(); }
    DataT C63() const { return C<6, 3>(); }
    DataT C64() const { return C<6, 4>(); }
    DataT C65() const { return C<6, 5>(); }
    DataT C66() const { return C<6, 6>(); }

    /// ---------------------------------------------------------------------------------------------------------------------
    ///  Getters with 'Get' prefix
    ///  Some of them involve calculations

    /// \brief Gets z position [cm]
    DataT GetZ() const { return fZ; }

    /// \brief Gets x position [cm]
    DataT GetX() const { return fX; }

    /// \brief Gets x position error [cm]
    DataT GetXError() const { return sqrt(C00()); }

    /// \brief Gets y position [cm]
    DataT GetY() const { return fY; }

    /// \brief Gets y position error [cm]
    DataT GetYError() const { return sqrt(C11()); }

    /// \brief Gets slope along x-axis
    DataT GetTx() const { return fTx; }

    /// \brief Gets error of slope along x-axis
    DataT GetTxError() const { return sqrt(C22()); }

    /// \brief Gets slope along y-axis
    DataT GetTy() const { return fTy; }

    /// \brief Gets error of slope along y-axis
    DataT GetTyError() const { return sqrt(C33()); }

    /// \brief Gets charge over momentum [ec/GeV]
    DataT GetQp() const { return fQp; }

    /// \brief Gets error of charge over momentum [ec/GeV]
    DataT GetQpError() const { return sqrt(C44()); }

    /// \brief Gets time [ns]
    DataT GetTime() const { return fT; }

    /// \brief Gets time error [ns]
    DataT GetTimeError() const { return sqrt(C55()); }

    /// \brief Gets inverse velocity [ns/cm] in downstream direction
    DataT GetVi() const { return fVi; }

    /// \brief Gets inverse velocity error [ns/cm]
    DataT GetViError() const { return sqrt(C66()); }

    /// \brief Gets covariance matrix
    const CovMatrix_t& GetCovMatrix() const { return fCovMatrix; }

    /// \brief Get covariance matrix element
    /// \param i row
    /// \param j column
    /// \return covariance matrix element
    DataT GetCovariance(int i, int j) const { return C(i, j); }

    /// Gets Chi-square of track fit model
    DataT GetChiSq() const { return fChiSq; }

    /// Gets NDF of track fit model
    DataT GetNdf() const { return fNdf; }

    /// Gets Chi-square of time measurements
    DataT GetChiSqTime() const { return fChiSqTime; }

    /// Gets NDF of time measurements
    DataT GetNdfTime() const { return fNdfTime; }

    /// Gets charge
    DataT GetCharge() const { return utils::iif(GetQp() > DataT(0.), DataT(1.), DataT(-1.)); }

    /// \brief Gets azimuthal angle [rad]
    DataT GetPhi() const { return atan2(GetTy(), GetTx()); }

    /// \brief Gets azimuthal angle error [rad]
    DataT GetPhiError() const;

    /// \brief Gets polar angle [rad]
    DataT GetTheta() const { return atan(sqrt(GetTx() * GetTx() + GetTy() * GetTy())); }

    /// \brief Gets polar angle error [rad]
    DataT GetThetaError() const;

    /// Gets momentum [GeV/ec]. For the straight tracks returns 1.e4 [GeV/ec]
    DataT GetP() const
    {
      return utils::iif(utils::fabs(GetQp()) > DataT(1.e-4), DataT(1.) / utils::fabs(GetQp()), DataT(1.e4));
    }

    /// Gets z-component of the momentum [GeV/ec]
    DataT GetPz() const { return GetP() / sqrt(DataT(1.) + GetTx() * GetTx() + GetTy() * GetTy()); }

    /// Gets x-component of the momentum [GeV/ec]
    DataT GetPx() const { return GetPz() * GetTx(); }

    /// Gets y-component of the momentum [GeV/ec]
    DataT GetPy() const { return GetPz() * GetTy(); }

    /// Gets transverse momentum
    DataT GetPt() const
    {
      DataT t2 = GetTx() * GetTx() + GetTy() * GetTy();
      return GetP() * sqrt(t2 / (DataT(1.) + t2));
    }

    /// ---------------------------------------------------------------------------------------------------------------------
    /// Parameter setters
    ///

    /// \brief Sets z position [cm]
    void SetZ(DataT v) { fZ = v; }

    /// \brief Sets x position [cm]
    void SetX(DataT v) { fX = v; }

    /// \brief Sets y position [cm]
    void SetY(DataT v) { fY = v; }

    /// \brief Sets slope along x-axis
    void SetTx(DataT v) { fTx = v; }

    /// \brief Sets slope along y-axis
    void SetTy(DataT v) { fTy = v; }

    /// \brief Sets charge over momentum [ec/GeV]
    void SetQp(DataT v) { fQp = v; }

    /// \brief Sets time [ns]
    void SetTime(DataT v) { fT = v; }

    /// \brief Sets inverse velocity [ns/cm]
    void SetVi(DataT v) { fVi = v; }

    /// \brief Sets Chi-square of track fit model
    void SetChiSq(DataT v) { fChiSq = v; }

    /// \brief Sets NDF of track fit model
    void SetNdf(DataT v) { fNdf = v; }

    /// \brief Sets Chi-square of time measurements
    void SetChiSqTime(DataT v) { fChiSqTime = v; }

    /// \brief Sets NDF of time measurements
    void SetNdfTime(DataT v) { fNdfTime = v; }

    /// \brief Sets covariance matrix
    void SetCovMatrix(const CovMatrix_t& val) { fCovMatrix = val; }

    /// \brief Get covariance matrix element
    /// \param i row
    /// \param j column
    /// \param val covariance matrix element
    void SetCovariance(int i, int j, DataT val)
    {
      int ind         = (j <= i) ? i * (1 + i) / 2 + j : j * (1 + j) / 2 + i;
      fCovMatrix[ind] = val;
    }

    /// \brief Get covariance matrix element when indices are known at compile time
    /// \param i row
    /// \param j column
    /// \param val matrix element
    template<int i, int j>
    void SetCovariance(DataT val)
    {
      constexpr int ind = (j <= i) ? i * (1 + i) / 2 + j : j * (1 + j) / 2 + i;
      fCovMatrix[ind]   = val;
    }

    /// \brief Individual setters for covariance matrix elements
    ///
    void SetC00(DataT val) { SetCovariance<0, 0>(val); }
    void SetC10(DataT val) { SetCovariance<1, 0>(val); }
    void SetC11(DataT val) { SetCovariance<1, 1>(val); }
    void SetC20(DataT val) { SetCovariance<2, 0>(val); }
    void SetC21(DataT val) { SetCovariance<2, 1>(val); }
    void SetC22(DataT val) { SetCovariance<2, 2>(val); }
    void SetC30(DataT val) { SetCovariance<3, 0>(val); }
    void SetC31(DataT val) { SetCovariance<3, 1>(val); }
    void SetC32(DataT val) { SetCovariance<3, 2>(val); }
    void SetC33(DataT val) { SetCovariance<3, 3>(val); }
    void SetC40(DataT val) { SetCovariance<4, 0>(val); }
    void SetC41(DataT val) { SetCovariance<4, 1>(val); }
    void SetC42(DataT val) { SetCovariance<4, 2>(val); }
    void SetC43(DataT val) { SetCovariance<4, 3>(val); }
    void SetC44(DataT val) { SetCovariance<4, 4>(val); }
    void SetC50(DataT val) { SetCovariance<5, 0>(val); }
    void SetC51(DataT val) { SetCovariance<5, 1>(val); }
    void SetC52(DataT val) { SetCovariance<5, 2>(val); }
    void SetC53(DataT val) { SetCovariance<5, 3>(val); }
    void SetC54(DataT val) { SetCovariance<5, 4>(val); }
    void SetC55(DataT val) { SetCovariance<5, 5>(val); }
    void SetC60(DataT val) { SetCovariance<6, 0>(val); }
    void SetC61(DataT val) { SetCovariance<6, 1>(val); }
    void SetC62(DataT val) { SetCovariance<6, 2>(val); }
    void SetC63(DataT val) { SetCovariance<6, 3>(val); }
    void SetC64(DataT val) { SetCovariance<6, 4>(val); }
    void SetC65(DataT val) { SetCovariance<6, 5>(val); }
    void SetC66(DataT val) { SetCovariance<6, 6>(val); }

    ///---------------------------------------------------------------------------------------------------------------------
    /// References to parameters to ease the math formulae
    /// They are especially useful for masked SIMD operations like:  X()( mask ) = fvec(1.);
    ///

    /// \brief Reference to z position [cm]
    DataT& Z() { return fZ; }

    /// \brief Reference to x position [cm]
    DataT& X() { return fX; }

    /// \brief Reference to y position [cm]
    DataT& Y() { return fY; }

    /// \brief Reference to slope along x-axis
    DataT& Tx() { return fTx; }

    /// \brief Reference to slope along y-axis
    DataT& Ty() { return fTy; }

    /// \brief Reference to charge over momentum [ec/GeV]
    DataT& Qp() { return fQp; }

    /// \brief Reference to time [ns]
    DataT& Time() { return fT; }

    /// \brief Reference to inverse velocity [ns/cm]
    DataT& Vi() { return fVi; }

    /// \brief Reference to covariance matrix
    CovMatrix_t& CovMatrix() { return fCovMatrix; }

    /// \brief Reference to Chi-square of track fit model
    DataT& ChiSq() { return fChiSq; }

    /// \brief Reference to NDF of track fit model
    DataT& Ndf() { return fNdf; }

    /// \brief Reference to Chi-square of time measurements
    DataT& ChiSqTime() { return fChiSqTime; }

    /// \brief Reference to NDF of time measurements
    DataT& NdfTime() { return fNdfTime; }

    /// \brief Get a reference to the covariance matrix element
    /// \param i row
    /// \param j column
    /// \return matrix element
    DataT& C(int i, int j)
    {
      int ind = (j <= i) ? i * (1 + i) / 2 + j : j * (1 + j) / 2 + i;
      return fCovMatrix[ind];
    }

    /// \brief Get a reference to the covariance matrix element when indices are known at compile time
    /// \param i row
    /// \param j column
    /// \param val matrix element
    template<int i, int j>
    DataT& C()
    {
      constexpr int ind = (j <= i) ? i * (1 + i) / 2 + j : j * (1 + j) / 2 + i;
      return fCovMatrix[ind];
    }

    /// \brief Individual references to covariance matrix elements
    ///
    DataT& C00() { return C<0, 0>(); }
    DataT& C01() { return C<0, 1>(); }
    DataT& C02() { return C<0, 2>(); }
    DataT& C03() { return C<0, 3>(); }
    DataT& C04() { return C<0, 4>(); }
    DataT& C05() { return C<0, 5>(); }
    DataT& C06() { return C<0, 6>(); }

    DataT& C10() { return C<1, 0>(); }
    DataT& C11() { return C<1, 1>(); }
    DataT& C12() { return C<1, 2>(); }
    DataT& C13() { return C<1, 3>(); }
    DataT& C14() { return C<1, 4>(); }
    DataT& C15() { return C<1, 5>(); }
    DataT& C16() { return C<1, 6>(); }

    DataT& C20() { return C<2, 0>(); }
    DataT& C21() { return C<2, 1>(); }
    DataT& C22() { return C<2, 2>(); }
    DataT& C23() { return C<2, 3>(); }
    DataT& C24() { return C<2, 4>(); }
    DataT& C25() { return C<2, 5>(); }
    DataT& C26() { return C<2, 6>(); }

    DataT& C30() { return C<3, 0>(); }
    DataT& C31() { return C<3, 1>(); }
    DataT& C32() { return C<3, 2>(); }
    DataT& C33() { return C<3, 3>(); }
    DataT& C34() { return C<3, 4>(); }
    DataT& C35() { return C<3, 5>(); }
    DataT& C36() { return C<3, 6>(); }

    DataT& C40() { return C<4, 0>(); }
    DataT& C41() { return C<4, 1>(); }
    DataT& C42() { return C<4, 2>(); }
    DataT& C43() { return C<4, 3>(); }
    DataT& C44() { return C<4, 4>(); }
    DataT& C45() { return C<4, 5>(); }
    DataT& C46() { return C<4, 6>(); }

    DataT& C50() { return C<5, 0>(); }
    DataT& C51() { return C<5, 1>(); }
    DataT& C52() { return C<5, 2>(); }
    DataT& C53() { return C<5, 3>(); }
    DataT& C54() { return C<5, 4>(); }
    DataT& C55() { return C<5, 5>(); }
    DataT& C56() { return C<5, 6>(); }

    DataT& C60() { return C<6, 0>(); }
    DataT& C61() { return C<6, 1>(); }
    DataT& C62() { return C<6, 2>(); }
    DataT& C63() { return C<6, 3>(); }
    DataT& C64() { return C<6, 4>(); }
    DataT& C65() { return C<6, 5>(); }
    DataT& C66() { return C<6, 6>(); }


    ///---------------------------------------------------------------------------------------------------------------------
    /// Other methods

    /// \brief Resets variances of track parameters and chi2, ndf values
    /// \param c00  Variance of x-position [cm2]
    /// \param c11  Variance of y-position [cm2]
    /// \param c22  Variance of slope along x-axis
    /// \param c33  Variance of slope along y-axis
    /// \param c44  Variance of charge over momentum [(ec/GeV)2]
    /// \param c55  Variance of time [ns2]
    /// \param c66  Variance of inverse velocity [1/c2]
    void ResetErrors(DataT c00, DataT c11, DataT c22, DataT c33, DataT c44, DataT c55, DataT c66);

    /// \brief Prints parameters to a string
    /// \param i  Index of SIMD vector element (when i== -1, the entire vector is printed)
    std::string ToString(int i = -1) const;

    /// \brief Prints correlations to a string
    /// \param i  Index of SIMD vector element (when i== -1, the entire vector is printed)
    std::string ToStringCorrelations(int i = -1) const;

    /// \brief Checks whether some parameters are finite
    bool IsFinite(bool printWhenWrong) const;

    /// \brief Checks whether SIMD entry i is consistent
    bool IsEntryConsistent(bool printWhenWrong, int i) const;

    /// \brief Checks whether first nFilled SIMD entries are consistent
    bool IsConsistent(bool printWhenWrong, int nFilled) const;

    /// \brief Initializes inverse velocity range
    void InitVelocityRange(fscal minP);


    ///---------------------------------------------------------------------------------------------------------------------
    /// Serialization
    ///
    template<class Archive>
    void serialize(Archive& ar, const unsigned int /*version*/)
    {
      ar& fX;
      ar& fY;
      ar& fTx;
      ar& fTy;
      ar& fQp;
      ar& fZ;
      ar& fT;
      ar& fVi;
      ar& fCovMatrix;
      ar& fChiSq;
      ar& fNdf;
      ar& fChiSqTime;
      ar& fNdfTime;
    }

   private:
    /// \brief Copies all/one entries from the other class
    /// \tparam TdataB  Type of the other class
    /// \tparam TDoAllA  If true, all entries of the current class must be set
    /// \tparam TDoAllB  If true, all entries of the other class must be used
    /// \param ia  Index of SIMD vector element of the current class
    /// \param Tb  Other class
    /// \param ib  Index of SIMD vector element of the other class
    template<typename TdataB, bool TDoAllA, bool TDoAllB>
    void CopyBase(const int ia, const TrackParamBase<TdataB>& Tb, const int ib);


   private:
    /// ---------------------------------------------------------------------------------------------------------------------
    /// Class members

    // Initializing parameters with NANs spoils the track fit where
    // the masked-out SIMD entries are suppressed by a multication by 0.
    // Therefore, we initialize the data members here with finite numbers.
    // For the numerical safety, with some reasonable numbers.

    // clang-format off
    CovMatrix_t fCovMatrix {1.,
                            0., 1.,
                            0., 0., 1.,
                            0., 0., 0., 1.,
                            0., 0., 0., 0., 1.,
                            0., 0., 0., 0., 0., 1.,
                            0., 0., 0., 0., 0., 0., 1.};  ///< covariance matrix
    // clang-format on

    DataT fX{0.};   ///< x-position [cm]
    DataT fY{0.};   ///< y-position [cm]
    DataT fZ{0.};   ///< z-position [cm]
    DataT fTx{0.};  ///< slope along x-axis
    DataT fTy{0.};  ///< slope along y-axis
    DataT fQp{0.};  ///< charge over momentum [ec/GeV]
    DataT fT{0.};   ///< time [ns]
    DataT fVi{0.};  ///< inverse velocity in downstream direction [ns/cm]

    DataT fChiSq{0.};      ///< chi^2 of track fit, spatial measurements
    DataT fNdf{0.};        ///< NDF of track fit, spatial measurements
    DataT fChiSqTime{0.};  ///< chi^2 of track fit, time measurements
    DataT fNdfTime{0.};    ///< NDF of track fit, time measurements

  } _fvecalignment;  // class TrackParamBase


  /// \class cbm::algo::ca:: TrackParamBaseScalar
  /// \brief Scalar version of TrackParamBase
  ///
  /// It contains extra methods that are difficult to implement in SIMD version
  template<typename DataT>
  class TrackParamBaseScalar : public TrackParamBase<DataT> {
   public:
    /// \brief Gets pseudo-rapidity
    DataT GetEta() const { return -log(tan(this->GetTheta() * DataT(0.5))); }

  } _fvecalignment;

  // ---------------------------------------------------------------------------------------------------------------------
  /// TrackParam classes of different types

  typedef TrackParamBase<fvec> TrackParamV;
  typedef TrackParamBaseScalar<fscal> TrackParamS;
  typedef TrackParamBaseScalar<double> TrackParamD;


  /// ---------------------------------------------------------------------------------------------------------------------
  /// Inline and template function implementation


  /// ---------------------------------------------------------------------------------------------------------------------
  ///
  template<typename DataT>
  inline DataT TrackParamBase<DataT>::GetPhiError() const
  {
    // phi = atan( tx / ty ); }

    DataT phiDdenom = GetTx() * GetTx() + GetTy() * GetTy();
    DataT phiDTx    = -GetTy() / phiDdenom;  // partial derivative of phi over Tx
    DataT phiDTy    = +GetTx() / phiDdenom;  // partial derivative of phi over Ty

    DataT varTx   = C22();  // variance of Tx
    DataT varTy   = C33();  // variance of Ty
    DataT covTxTy = C32();  // covariance of Tx and Ty

    DataT varPhi = phiDTx * phiDTx * varTx + phiDTy * phiDTy * varTy + DataT(2.) * phiDTx * phiDTy * covTxTy;
    return sqrt(varPhi);
  }


  /// ---------------------------------------------------------------------------------------------------------------------
  ///
  template<typename DataT>
  inline DataT TrackParamBase<DataT>::GetThetaError() const
  {
    // theta = atan(sqrt( tx * tx + ty * ty) )

    DataT sumSqSlopes = GetTx() * GetTx() + GetTy() * GetTy();
    DataT thetaDdenom = sqrt(sumSqSlopes) * (DataT(1.) + sumSqSlopes);
    DataT thetaDTx    = GetTx() / thetaDdenom;
    DataT thetaDTy    = GetTy() / thetaDdenom;

    DataT varTx   = C22();  // variance of Tx
    DataT varTy   = C33();  // variance of Ty
    DataT covTxTy = C32();  // covariance of Tx and Ty

    DataT varTheta =
      thetaDTx * thetaDTx * varTx + thetaDTy * thetaDTy * varTy + DataT(2.) * thetaDTx * thetaDTy * covTxTy;

    return sqrt(varTheta);
  }


  // ---------------------------------------------------------------------------------------------------------------------
  //
  template<typename TdataA>
  template<typename TdataB, bool TDoAllA, bool TDoAllB>
  inline void TrackParamBase<TdataA>::CopyBase(const int ia, const TrackParamBase<TdataB>& Tb, const int ib)
  {
    auto copy = [&](TdataA& a, const TdataB& b) {
      utils::VecCopy<TdataA, TdataB, TDoAllA, TDoAllB>::CopyEntries(a, ia, b, ib);
    };

    copy(fX, Tb.GetX());
    copy(fY, Tb.GetY());
    copy(fTx, Tb.GetTx());
    copy(fTy, Tb.GetTy());
    copy(fQp, Tb.GetQp());
    copy(fZ, Tb.GetZ());
    copy(fT, Tb.GetTime());
    copy(fVi, Tb.GetVi());

    for (int i = 0; i < kNcovParam; ++i) {
      copy(fCovMatrix[i], Tb.GetCovMatrix()[i]);
    }

    copy(fChiSq, Tb.GetChiSq());
    copy(fNdf, Tb.GetNdf());
    copy(fChiSqTime, Tb.GetChiSqTime());
    copy(fNdfTime, Tb.GetNdfTime());

  }  // CopyBase


  // ---------------------------------------------------------------------------------------------------------------------
  //
  template<typename DataT>
  inline void TrackParamBase<DataT>::ResetErrors(DataT c00, DataT c11, DataT c22, DataT c33, DataT c44, DataT c55,
                                                 DataT c66)
  {
    fCovMatrix.fill(0.);

    SetC00(c00);
    SetC11(c11);
    SetC22(c22);
    SetC33(c33);
    SetC44(c44);
    SetC55(c55);
    SetC66(c66);

    SetChiSq(0.);
    SetNdf(-5.);
    SetChiSqTime(0.);
    SetNdfTime(-2.);
  }

  // ---------------------------------------------------------------------------------------------------------------------
  //
  template<typename DataT>
  inline void TrackParamBase<DataT>::InitVelocityRange(fscal minP)
  {
    // initialise the velocity range with respect to the minimal momentum minP {GeV/c}
    namespace phys = constants::phys;
    fscal maxVi    = sqrt(1. + (phys::ProtonMassD / minP) * (phys::ProtonMassD / minP)) * phys::SpeedOfLightInvD;
    fscal minVi    = phys::SpeedOfLightInvD;
    fscal vmean    = minVi + 0.4 * (maxVi - minVi);
    fscal dvi      = (maxVi - vmean) / 3.;
    SetVi(vmean);
    SetC66(dvi * dvi);
  }

}  // namespace cbm::algo::ca

#endif  // CA_CORE_CaTrackParam_h
