/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// @file   KfDefs.h
/// @brief  Common constant definitions for the Kalman Filter library
/// @since  28.03.2024
/// @author Sergei Zharko <s.zharko@gsi.de>

#pragma once

#include <limits>

namespace cbm::algo::kf::defs
{
  // ----- Control -----------------------------------------------------------------------------------------------------
  constexpr int DebugLvl     = 0;     ///< Level of debug output
  constexpr bool GetterCheck = true;  ///< Bound check in getters


  // ----- Math --------------------------------------------------------------------------------------------------------
  template<class T = double>
  constexpr T Pi = T(3.14159265358979323846L);  ///< Value of PI, used in ROOT TMath


  // ----- Physics -----------------------------------------------------------------------------------------------------
  template<class T = double>
  constexpr T MuonMass = T(0.105658375523L);  ///< Muon mass [GeV/c2]

  template<class T = double>
  constexpr T PionMass = T(0.1395703918L);  ///< Pion mass [GeV/c2]

  template<class T = double>
  constexpr T KaonMass = T(0.493677L);  ///< Kaon mass [GeV/c2] (PDG 22.08.2023)

  template<class T = double>
  constexpr T ElectronMass = T(0.0005109989500015L);  ///< Electron mass [GeV/c2]

  template<class T = double>
  constexpr T ProtonMass = T(0.93827208816L);  ///< Proton mass [GeV/c2] (PDG 11.08.2022)

  template<class T = double>
  constexpr T SpeedOfLight = T(29.9792458L);  ///< Speed of light [cm/ns]

  template<class T = double>
  constexpr T SpeedOfLightInv = T(1.L) / SpeedOfLight<T>;  ///< Inverse speed of light [ns/cm]

  template<class T = double>
  constexpr T MinFeild = T(1.e-4L);  ///< Minimal (negligible) magnetic field value [kG]


  // ----- Misc --------------------------------------------------------------------------------------------------------
  constexpr int MemAlign = 16;  ///< Default alignment of memory (bytes)


  // ----- Undefined values --------------------------------------------------------------------------------------------
  /// \brief Undefined values
  template<typename T1, typename T2 = T1>
  constexpr T2 Undef;

  template<>
  inline constexpr int Undef<int> = std::numeric_limits<int>::min();

  template<>
  inline constexpr unsigned Undef<unsigned> = std::numeric_limits<unsigned>::max();

  template<>
  inline constexpr float Undef<float> = std::numeric_limits<float>::signaling_NaN();

  template<>
  inline constexpr double Undef<double> = std::numeric_limits<double>::signaling_NaN();

  // TODO: provide SIMD
  //template<>
  //inline constexpr fscal Undef<fvec> = std::numeric_limits<fscal>::signaling_NaN();

}  // namespace cbm::algo::kf::defs
