/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// @file   KfFramework.h
/// @brief  The Kalman-filter framework main class (header)
/// @since  28.03.2024
/// @author Sergei Zharko <s.zharko@gsi.de>

#pragma once

#include "KfDefs.h"
#include "KfParameters.h"
#include "KfSetup.h"

namespace cbm::algo::kf
{
  /// \class Framework
  /// \brief Main class of the KfCore library
  /// \tparam DataT  Underlying data-type
  template<typename DataT>
  class Framework {
   public:
    /// \brief Default constructor
    Framework() = default;

    /// \brief Copy constructor
    Framework(const Framework&) = delete;

    /// \brief Move constructor
    Framework(Framework&&) = delete;

    /// \brief Destructor
    ~Framework() = default;

    /// \brief Copy assignment operator
    Framework& operator=(const Framework&) = delete;

    /// \brief Move assignment operator
    Framework& operator=(Framework&&) = delete;

    /// \brief Initialization method
    bool Init();

    /// \brief Initialization satus
    bool IsInitialized() const { return fbInitialized; }

    /// \brief Parameters access
    Parameters<DataT>& Pars() { return fPars; }

    /// \brief Parameters access
    const kf::Parameters<DataT>& Pars() const { return fPars; }

    /// \brief Setup access (mutable)
    kf::Setup<DataT>& Setup() { return fSetup; }

    /// \brief Setup access
    const kf::Setup<DataT>& Setup() const { return fSetup; }

   private:
    kf::Parameters<DataT> fPars;  ///< KF parameters
    kf::Setup<DataT> fSetup;      ///< KF setup
    bool fbInitialized = false;   ///< Initialization status
  };
}  // namespace cbm::algo::kf
