/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CbmQaOnlineInterface.h
/// \date   28.02.2024
/// \brief  Set of tools for online->ROOT QA-objects conversions (header)
/// \author Sergei Zharko <s.zharko@gsi.de>

// TODO: move to more suitable place (histoserv, for example)

#pragma once

#include "Histogram.h"
#include "Logger.h"
#include "TH1D.h"
#include "TH2D.h"
#include "TProfile.h"
#include "TProfile2D.h"

#include <log.hpp>
#include <type_traits>


namespace cbm::qa
{
  /// \class RootHistogramAccessor
  /// \brief Helper class to access protected fields of TH1, TH2, TProfile and TProfile2D
  ///
  /// The ROOT TProfile classes do not alow to set a profile from outside, so we need to set all the
  /// fields directly. Also the access to TH1/TH2 protected fields is needed in order to set the
  /// total sums of w*x, w*x*x, w*y, w*x*y, and w*y*y.
  template<class RootHistogram, class QaHistogram>
  class RootHistogramAccessor : public RootHistogram {

   public:
    template<class... Args>
    RootHistogramAccessor(Args... args) : RootHistogram(args...)
    {
    }

    /// \brief Sets fields from qa histogram
    void SetFromQaHistogram(const QaHistogram& histo);
  };

  /// \class  OnlineInterface
  /// \brief  A collection of tools for online QA object conversions
  class OnlineInterface {
   public:
    /// \brief Converts histogram H1D to ROOT histogram TH1D
    /// \param hist  1D-histogram
    /// \note  Allocates memory on heap for the ROOT histogram
    static TH1D* ROOTHistogram(const cbm::algo::qa::H1D& hist);

    /// \brief Converts histogram H2D to ROOT histogram TH2D
    /// \param hist  1D-histogram
    /// \note  Allocates memory on heap for the ROOT histogram
    static TH2D* ROOTHistogram(const cbm::algo::qa::H2D& hist);

    /// \brief Converts profile Prof1D to ROOT TProfile
    /// \param prof  1D-profile
    /// \note  Allocates memory on heap for the ROOT histogram
    static TProfile* ROOTHistogram(const cbm::algo::qa::Prof1D& prof);

    /// \brief Converts profile Prof2D to ROOT TProfile2D
    /// \param prof  2D-profile
    /// \note  Allocates memory on heap for the ROOT histogram
    static TProfile2D* ROOTHistogram(const cbm::algo::qa::Prof2D& prof);
  };
}  // namespace cbm::qa


namespace cbm::qa
{
  // -------------------------------------------------------------------------------------------------------------------
  //
  template<class RootHistogram, class QaHistogram>
  void RootHistogramAccessor<RootHistogram, QaHistogram>::SetFromQaHistogram(const QaHistogram& hist)
  {
    // Set individual bin properties
    if constexpr (std::is_same_v<RootHistogram, TProfile> || std::is_same_v<RootHistogram, TProfile2D>) {
      this->Sumw2();
      auto SetBin = [&](int iBinRoot, const auto& binQa) {
        this->fArray[iBinRoot]             = binQa.GetSumWV();
        this->fBinEntries.fArray[iBinRoot] = binQa.GetSumW();
        this->fSumw2.fArray[iBinRoot]      = binQa.GetSumWV2();
        this->fBinSumw2.fArray[iBinRoot]   = binQa.GetSumW2();
      };
      if constexpr (std::is_same_v<RootHistogram, TProfile>) {
        for (int iBinX = 0; iBinX <= this->GetNbinsX() + 1; ++iBinX) {
          int iBin = this->GetBin(iBinX);
          auto bin = hist.GetBinAccumulator(iBinX);
          SetBin(iBin, bin);
          this->fTsumwy += bin.GetSumWV();
          this->fTsumwy2 += bin.GetSumWV2();
        }
      }
      else if constexpr (std::is_same_v<RootHistogram, TProfile2D>) {
        for (int iBinX = 0; iBinX <= this->GetNbinsX() + 1; ++iBinX) {
          for (int iBinY = 0; iBinY <= this->GetNbinsY() + 1; ++iBinY) {
            int iBin = this->GetBin(iBinX, iBinY);
            auto bin = hist.GetBinAccumulator(iBinX, iBinY);
            SetBin(iBin, bin);
            this->fTsumwz += bin.GetSumWV();
            this->fTsumwz2 += bin.GetSumWV2();
          }
        }
      }
    }
    else if constexpr (std::is_same_v<RootHistogram, TH1D> || std::is_same_v<RootHistogram, TH2D>) {
      if (hist.GetTotSumW() != hist.GetTotSumW2()) {  // some of weights were not equal to 1.
        if (!this->fSumw2.fN) {
          this->Sumw2();
        }
      }
      auto SetBin = [&](int iBinRoot, const auto& binQa) {
        this->fArray[iBinRoot] = binQa.value();
        if (this->fSumw2.fN) {
          this->fSumw2.fArray[iBinRoot] = binQa.variance();
        }
      };
      if constexpr (std::is_same_v<RootHistogram, TH1D>) {
        for (int iBinX = 0; iBinX <= this->GetNbinsX() + 1; ++iBinX) {
          SetBin(this->GetBin(iBinX), hist.GetBinAccumulator(iBinX));
        }
      }
      else if constexpr (std::is_same_v<RootHistogram, TH2D>) {
        for (int iBinX = 0; iBinX <= this->GetNbinsX() + 1; ++iBinX) {
          for (int iBinY = 0; iBinY <= this->GetNbinsY() + 1; ++iBinY) {
            SetBin(this->GetBin(iBinX, iBinY), hist.GetBinAccumulator(iBinX, iBinY));
          }
        }
      }
    }

    // Set total sums
    if constexpr (std::is_base_of_v<TH2D, RootHistogram>) {
      this->fTsumwy  = hist.GetTotSumWY();
      this->fTsumwy2 = hist.GetTotSumWY2();
      this->fTsumwxy = hist.GetTotSumWXY();
    }
    this->fTsumw   = hist.GetTotSumW();
    this->fTsumw2  = hist.GetTotSumW2();
    this->fTsumwx  = hist.GetTotSumWX();
    this->fTsumwx2 = hist.GetTotSumWX2();
    this->SetEntries(hist.GetEntries());
  }
}  // namespace cbm::qa
