/* Copyright (C) 2021-2023 UGiessen, JINR-LIT
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Semen Lebedev [committer], Martin Beyer */

#if !defined(__CLING__) || defined(__ROOTCLING__)
#include "CbmTransport.h"

#include <FairBoxGenerator.h>
#include <FairLogger.h>
#include <FairRunSim.h>

#include <TStopwatch.h>
#include <TTree.h>

#include <iostream>
#endif

void run_transport(TString urqmdFile = "default10gev",  // if "", no urqmd; "default10gev" or "default25gev"
                   TString traFile = "", TString parFile = "", TString geoFile = "",  // i/o files
                   Int_t nEvents     = 1,
                   Int_t nElectrons  = 5,   // number of embedded e- into each event, generated by the BoxGenerator
                   Int_t nPositrons  = 5,   // number of embedded e+ into each event, generated by the BoxGenerator
                   TString plutoFile = "",  // if "", no pluto particles are embedded into event
                   TString geoSetup = "sis100_electron", Int_t engine = 0,  // geant3: 0, geant4: 1
                   Int_t randomSeed = 0)
{
  TTree::SetMaxTreeSize(90000000000);

  Double_t targetZ = -44.0;

  // -----   Files   --------------------------------------------------------
  TString macroPath = __FILE__;
  TString macroDir  = macroPath(0, macroPath.Last('/') + 1);
  if (traFile.IsNull()) traFile = macroDir + "data/test.tra.root";
  if (parFile.IsNull()) parFile = macroDir + "data/test.par.root";
  if (geoFile.IsNull()) geoFile = macroDir + "data/test.geo.root";

  remove(traFile);
  remove(parFile);
  remove(geoFile);
  // ------------------------------------------------------------------------

  // -----   Logger settings   ----------------------------------------------
  FairLogger::GetLogger()->SetLogScreenLevel("INFO");
  FairLogger::GetLogger()->SetLogVerbosityLevel("LOW");
  // ------------------------------------------------------------------------

  // -----   Environment   --------------------------------------------------
  TString srcDir            = gSystem->Getenv("VMCWORKDIR");
  TString defaultInput10gev = srcDir + "/input/urqmd.auau.10gev.centr.root";
  TString defaultInput25gev = srcDir + "/input/urqmd.auau.25gev.centr.root";
  // ------------------------------------------------------------------------

  // -----   Timer   --------------------------------------------------------
  TStopwatch timer;
  timer.Start();
  // ------------------------------------------------------------------------

  // --- Transport run   ----------------------------------------------------
  CbmTransport run;
  run.SetEngine(static_cast<ECbmEngine>(engine));

  if (nElectrons > 0) {
    FairBoxGenerator* boxGen1 = new FairBoxGenerator(11, nElectrons);
    boxGen1->SetPtRange(0., 3.);
    // boxGen1->SetPRange(0., 3.);
    boxGen1->SetPhiRange(0., 360.);
    boxGen1->SetThetaRange(2.5, 25.);
    // boxGen1->SetXYZ(0., 0., targetZ); // Dont use together with run.SetTarget(..., targetZ), boxGen will be at 2*targetZ
    boxGen1->SetCosTheta();
    boxGen1->Init();
    run.AddInput(boxGen1);
  }
  if (nPositrons > 0) {
    FairBoxGenerator* boxGen2 = new FairBoxGenerator(-11, nPositrons);
    boxGen2->SetPtRange(0., 3.);
    // boxGen2->SetPRange(0., 3.);
    boxGen2->SetPhiRange(0., 360.);
    boxGen2->SetThetaRange(2.5, 25.);
    // boxGen2->SetXYZ(0., 0., targetZ); // Dont use together with run.SetTarget(..., targetZ), boxGen will be at 2*targetZ
    boxGen2->SetCosTheta();
    boxGen2->Init();
    run.AddInput(boxGen2);
  }

  if (!plutoFile.IsNull()) {
    run.AddInput(plutoFile, kPluto);
  }

  if (!urqmdFile.IsNull()) {
    if (urqmdFile == "default10gev") {
      run.AddInput(defaultInput10gev);
    }
    else if (urqmdFile == "default25gev") {
      run.AddInput(defaultInput25gev);
    }
    else {
      run.AddInput(urqmdFile);
    }
  }

  run.SetOutFileName(traFile);
  run.SetParFileName(parFile);
  run.SetGeoFileName(geoFile);
  run.LoadSetup(geoSetup);
  run.SetTarget("Gold", 0.025, 2.5, 0, 0, targetZ);
  run.SetBeamPosition(0., 0., 0.1, 0.1);
  run.SetRandomEventPlane();
  run.SetRandomSeed(randomSeed);
  // run.StoreTrajectories(true);
  run.Run(nEvents);
  // ------------------------------------------------------------------------

  // -----   Finish   -------------------------------------------------------
  timer.Stop();
  std::cout << std::endl;
  std::cout << "Macro finished successfully." << std::endl;
  std::cout << "Output file is " << traFile << std::endl;
  std::cout << "Parameter file is " << parFile << std::endl;
  std::cout << "Geometry file is " << geoFile << std::endl;
  std::cout << "Real time " << timer.RealTime() << " s, CPU time " << timer.CpuTime() << "s" << std::endl;
  std::cout << std::endl << "Test passed" << std::endl << "All ok" << std::endl;
  // ------------------------------------------------------------------------
}
