/* Copyright (C) 2024 Hulubei National Institute of Physics and Nuclear Engineering - Horia Hulubei, Bucharest
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Alexandru Bercuci [committer] */

// CBM headers
#include "CbmRecoQaTask.h"

#include "../../core/detectors/rich/utils/CbmRichUtil.h"
#include "CbmDefs.h"
#include "CbmEvent.h"
#include "CbmFsdHit.h"
#include "CbmGlobalTrack.h"
#include "CbmMvdHit.h"
#include "CbmPsdHit.h"
#include "CbmRichHit.h"
#include "CbmSetup.h"
#include "CbmStsAddress.h"
#include "CbmStsHit.h"
#include "CbmTimeSlice.h"
#include "CbmTofAddress.h"
#include "CbmTofHit.h"
#include "CbmTrdAddress.h"
#include "CbmTrdHit.h"
// FAIR headers
#include <FairRootManager.h>
#include <FairSink.h>
// ROOT headers
#include <TClonesArray.h>
#include <TGeoManager.h>
#include <TGeoNode.h>
#include <TH2D.h>

using namespace std;
std::bitset<kRecoQaNConfigs> CbmRecoQaTask::fuRecoConfig = {};

//_____________________________________________________________________
CbmRecoQaTask::CbmRecoQaTask() : FairTask("RecoQA", 0) {}

//_____________________________________________________________________
CbmRecoQaTask::Detector* CbmRecoQaTask::AddDetector(ECbmModuleId id)
{
  /* Interface to the user. Check that the det defined in the outside world fulfills the quality criteria. 
  */
  switch (id) {
    case ECbmModuleId::kMvd:
    case ECbmModuleId::kSts:
    case ECbmModuleId::kMuch:
    case ECbmModuleId::kRich:
    case ECbmModuleId::kTrd:
    case ECbmModuleId::kTrd2d:
    case ECbmModuleId::kTof:
    case ECbmModuleId::kFsd:
    case ECbmModuleId::kPsd:
      if (GetDetector(id)) {
        LOG(warn) << GetName() << "::AddDetector : det " << ToString(id)
                  << " already registered. Using \"CbmRecoQaTask::GetDetector(ECbmModuleId).\"";
        return GetDetector(id);
      }
      LOG(debug) << GetName() << "::AddDetector(" << ToString(id) << ").";
      fDetQa.emplace(id, id);
      break;
    default: LOG(warn) << GetName() << "::AddDetector : unsupported det " << ToString(id); return nullptr;
  }

  return &fDetQa[id];
}

//_____________________________________________________________________
CbmRecoQaTask::Detector* CbmRecoQaTask::GetDetector(ECbmModuleId id)
{
  if (fDetQa.find(id) == fDetQa.end()) return nullptr;
  return &fDetQa[id];
}

//_____________________________________________________________________
const CbmTrack* CbmRecoQaTask::GetTrack(ECbmModuleId did, int id) const
{
  if (fTracks.find(did) == fTracks.end()) {
    LOG(warning) << GetName() << " missing tracks for " << ToString(did);
    return nullptr;
  }
  if (id < 0) {
    LOG(warning) << GetName() << " negative trk idx for " << ToString(did);
    return nullptr;
  }
  const CbmTrack* trk = (const CbmTrack*) (fTracks.at(did))->At(id);
  if (!trk) {
    LOG(debug1) << GetName() << " missing trk_" << id << " for " << ToString(did);
    return nullptr;
  }
  return trk;
}

//_____________________________________________________________________
InitStatus CbmRecoQaTask::Init()
{
  fFitter.Init();
  fFitter.SetSkipUnmeasuredCoordinates(true);

  //Get ROOT Manager
  FairRootManager* ioman = FairRootManager::Instance();

  if (!ioman) {
    LOG(fatal) << GetName() << "::Init :: RootManager not instantiated!";
    return kERROR;
  }

  fGTracks = static_cast<TClonesArray*>(ioman->GetObject("GlobalTrack"));
  if (!fGTracks)
    LOG(warn) << GetName() << "::Init: Global track array not found!";
  else
    fuRecoConfig.set(kRecoTracks);

  fEvents = static_cast<TClonesArray*>(ioman->GetObject("CbmEvent"));
  if (!fEvents)
    LOG(warn) << GetName() << "::Init: No event found. Some results will be missing.";
  else
    fuRecoConfig.set(kRecoEvents);

  fTrackMatches = static_cast<TClonesArray*>(ioman->GetObject("GlobalTrackMatch"));
  if (!fTrackMatches) {
    LOG(warn) << GetName() << "::Init: MC info for Global track not available !";
  }

  LOG(info) << GetName() << "::Init: Registered projections " << fProjs.size();

  switch (fSetupClass) {
    case kMcbm22:
      LOG(info) << GetName() << "::Init: Setup config for \"mCBM 2022\".";
      InitMcbm22();
      break;
    case kMcbm24:
      LOG(info) << GetName() << "::Init: Setup config for \"mCBM 2024\". ";
      InitMcbm24();
      break;
    case kDefault: LOG(info) << GetName() << "::Init: Setup config for \"DEFAULT\". Not implemented."; break;
  }

  fOutFolder.Clear();
  for (auto& detp : fDetQa) {  // SYSTEM LOOP
    auto& det = detp.second;

    fHits[det.id] = static_cast<TClonesArray*>(ioman->GetObject(det.hit.name.data()));
    if (!fHits[det.id]) {
      LOG(warn) << GetName() << "::Init: Hit array for " << ToString(det.id) << " not found!";
      continue;
    }

    if (fGTracks && det.trk.id != ECbmDataType::kUnknown) {
      fTracks[det.id] = static_cast<TClonesArray*>(ioman->GetObject(det.trk.name.data()));
      if (!fTracks[det.id]) LOG(warn) << GetName() << "::Init: Track array for " << ToString(det.id) << " not found!";
    }
    if (!det.Init(&fOutFolder)) continue;
  }

  // TDirectoryFile* trkDir = (TDirectoryFile*)fOutFolder.mkdir("track", "CA QA");
  // for (auto& trkly : fTrkView) {
  //   for (auto& view : trkly.second) view.Register(trkDir);
  // }

  TDirectoryFile* trgDir = (TDirectoryFile*) fOutFolder.mkdir("target", "Target tomography with CA");
  for (auto& prj : fProjs)
    prj.second.Register(trgDir);

  return kSUCCESS;
}

/// Hit classification on system and view
template<class Hit>
bool CbmRecoQaTask::Detector::View::HasAddress(const CbmHit* h, double&, double&) const
{
  LOG(error) << "Unprocessed hit in view " << name;
  cout << h->ToString();

  return false;
}
// STS specialization
template<>
bool CbmRecoQaTask::Detector::View::HasAddress<CbmStsHit>(const CbmHit* h, double& x, double& y) const
{
  int32_t a      = h->GetAddress();
  uint32_t sel[] = {CbmStsAddress::GetElementId(a, EStsElementLevel::kStsUnit),
                    CbmStsAddress::GetElementId(a, EStsElementLevel::kStsLadder),
                    CbmStsAddress::GetElementId(a, EStsElementLevel::kStsModule)};
  uint idx(0);
  for (auto ii : fSelector) {
    if (uint(ii) != sel[idx]) break;
    idx++;
  }
  if (idx != fSelector.size())
    return false;
  else
    LOG(debug4) << "Accept Sts hit for " << sel[0] << " " << sel[1] << " " << sel[2];

  const CbmStsHit* h0 = dynamic_cast<const CbmStsHit*>(h);
  if (!h0) {
    LOG(error) << "Failed loading STS hit in view " << name;
    return false;
  }
  x = h0->GetX();
  y = h0->GetY();
  return true;
}
// Rich specialization
template<>
bool CbmRecoQaTask::Detector::View::HasAddress<CbmRichHit>(const CbmHit* h, double& x, double& y) const
{
  int16_t uId = CbmRichUtil::GetDirichId(h->GetAddress());
  // TODO implement at RichUtil level
  int8_t modId = (uId >> 8) & 0xF;
  if (modId >= fSelector[0]) return false;

  const CbmRichHit* h0 = dynamic_cast<const CbmRichHit*>(h);
  if (!h0) {
    LOG(error) << "Failed loading RICH hit in view " << name;
    return false;
  }
  x = h0->GetX();
  y = h0->GetY();
  return true;
}
// TRD specialization
template<>
bool CbmRecoQaTask::Detector::View::HasAddress<CbmTrdHit>(const CbmHit* h, double& x, double& y) const
{
  int16_t uId = CbmTrdAddress::GetModuleAddress(h->GetAddress());
  if (uId != fSelector[0])
    return false;
  else
    LOG(debug4) << "Accept Trd hit for " << uId;


  const CbmTrdHit* h0 = dynamic_cast<const CbmTrdHit*>(h);
  if (!h0) {
    LOG(error) << "Failed loading TRD hit in view " << name;
    return false;
  }
  x = h0->GetX();
  y = h0->GetY();
  return true;
}
// ToF specialization
template<>
bool CbmRecoQaTask::Detector::View::HasAddress<CbmTofHit>(const CbmHit* h, double& x, double& y) const
{
  int32_t a     = h->GetAddress();
  int32_t sel[] = {CbmTofAddress::GetSmId(a), CbmTofAddress::GetSmType(a), CbmTofAddress::GetRpcId(a)};
  uint idx(0);
  for (auto ii : fSelector) {
    if (ii != sel[idx]) break;
    idx++;
  }
  if (idx != fSelector.size())
    return false;
  else
    LOG(debug4) << "Accept Tof hit for " << sel[0] << " " << sel[1] << " " << sel[2];

  const CbmTofHit* h0 = dynamic_cast<const CbmTofHit*>(h);
  if (!h0) {
    LOG(error) << "Failed loading ToF hit in view " << name;
    return false;
  }
  x = h0->GetX();
  y = h0->GetY();
  return true;
}
template<class Hit>
bool CbmRecoQaTask::Detector::View::Load(const CbmHit* h, const CbmEvent* ev)
{
  double x(0), y(0);
  if (!HasAddress<Hit>(h, x, y)) return false;

  //printf("View[%s] z(%d)[%f]\n", name.data(), (int)h->GetType(), h->GetZ());
  int scale(0);
  TH2* hh(nullptr);
  if (ev) {
    if (fProjection.find(eViewProjection::kChdT) != fProjection.end()) {
      scale = get<0>(fProjection[eViewProjection::kChdT]);
      hh    = get<2>(fProjection[eViewProjection::kChdT]);
      hh->Fill(x, scale * (h->GetTime() - ev->GetStartTime()));
    }
    if (fProjection.find(eViewProjection::kXYh) != fProjection.end()) {
      hh = get<2>(fProjection[eViewProjection::kXYh]);
      hh->Fill(x, y);
    }
  }
  if (fProjection.find(eViewProjection::kXYs) != fProjection.end()) {
    hh = get<2>(fProjection[eViewProjection::kXYs]);
    hh->Fill(x, y);
  }
  return true;
}

bool CbmRecoQaTask::Detector::View::Load(const CbmKfTrackFitter::FitNode* n)
{
  double dx = n->fMxy.X()[0] - n->fTrack.X()[0], dy = n->fMxy.Y()[0] - n->fTrack.Y()[0],
         dt    = n->fMt.T()[0] - n->fTrack.Time()[0],
         pullx = dx / sqrt(n->fMxy.Dx2()[0] + n->fTrack.GetCovariance(0, 0)[0]),
         pully = dy / sqrt(n->fMxy.Dy2()[0] + n->fTrack.GetCovariance(1, 1)[0]);
  // pullt = dt / sqrt(n->fMt.Dt2()[0] + n->fTrack.GetCovariance(5, 5)[0]);

  for (auto& projection : fProjection) {
    int scale = get<0>(projection.second);
    TH2* hh   = get<2>(projection.second);
    if (!hh) continue;

    switch (projection.first) {
      case eViewProjection::kXYa: hh->Fill(n->fMxy.X()[0], n->fMxy.Y()[0]); break;
      case eViewProjection::kXYp: hh->Fill(n->fTrack.X()[0], n->fTrack.Y()[0]); break;
      case eViewProjection::kXdX: hh->Fill(n->fMxy.X()[0], scale * dx); break;
      case eViewProjection::kYdY: hh->Fill(n->fMxy.Y()[0], scale * dy); break;
      case eViewProjection::kWdT: hh->Fill(n->fMxy.X()[0], scale * dt); break;
      case eViewProjection::kXpX: hh->Fill(n->fMxy.X()[0], pullx); break;
      case eViewProjection::kYpY: hh->Fill(n->fMxy.Y()[0], pully); break;
      default: break;
    }
  }
  return true;
}

//_____________________________________________________________________
void CbmRecoQaTask::Exec(Option_t*)
{
  LOG(info) << GetName() << "::Exec : Evs[" << (fEvents ? fEvents->GetEntriesFast() : 0) << "] Trks["
            << (fGTracks ? fGTracks->GetEntriesFast() : 0) << "]";

  int iev(0), itrack(0), nnodes(0);
  if (fEvents) {
    for (auto evObj : *fEvents) {  // EVENT LOOP
      auto ev = dynamic_cast<CbmEvent*>(evObj);
      // track QA filtering
      if (!FilterEvent(ev)) continue;

      for (auto detp : fDetQa) {  // SYSTEM LOOP
        auto det = detp.second;
        // check SYSTEM @ hit level
        int nh = max(int(0), int(ev->GetNofData(det.hit.id)));

        if (!fHits[det.id]) {
          LOG(error) << GetName() << "::Exec() : Hits for " << ToString(det.id) << " not available. Skip.";
          continue;
        }
        for (int ih(0); ih < nh; ih++) {  // HIT LOOP
          const int jh      = ev->GetIndex(det.hit.id, ih);
          const CbmHit* hit = dynamic_cast<CbmHit*>(fHits[det.id]->At(jh));
          // printf("%s hit_%2d dt=%f time_event=%f time_hit=%f\n", ToString(det.id).data(), ih, ev->GetStartTime() - hit->GetTime(), ev->GetStartTime(), hit->GetTime());

          bool ret(false);
          for (auto view : det.fViews) {  // VIEW LOOP
            switch (det.id) {
              case ECbmModuleId::kMvd: ret = view.Load<CbmMvdHit>(hit, ev); break;
              case ECbmModuleId::kSts: ret = view.Load<CbmStsHit>(hit, ev); break;
              case ECbmModuleId::kMuch: ret = view.Load<CbmMuchPixelHit>(hit, ev); break;
              case ECbmModuleId::kRich: ret = view.Load<CbmRichHit>(hit, ev); break;
              case ECbmModuleId::kTrd: ret = view.Load<CbmTrdHit>(hit, ev); break;
              case ECbmModuleId::kTrd2d: ret = view.Load<CbmTrdHit>(hit, ev); break;
              case ECbmModuleId::kTof: ret = view.Load<CbmTofHit>(hit, ev); break;
              case ECbmModuleId::kFsd: ret = view.Load<CbmFsdHit>(hit, ev); break;
              case ECbmModuleId::kPsd: ret = view.Load<CbmPsdHit>(hit, ev); break;
              default: LOG(fatal) << GetName() << "::Exec : unsupported det " << ToString(det.id); break;
            }
            if (ret) break;
          }  // END VIEW LOOP
        }    // END HIT LOOP
      }      // END SYSTEM LOOP

      if (!fGTracks) continue;

      int ntrk = max(int(0), int(ev->GetNofData(ECbmDataType::kGlobalTrack)));
      for (int itrk(0); itrk < ntrk; itrk++) {  // TRACK LOOP
        auto trackObj = (CbmGlobalTrack*) (*fGTracks)[ev->GetIndex(ECbmDataType::kGlobalTrack, itrk)];

        // track QA filtering
        if (!FilterTrack(trackObj)) continue;

        auto track = dynamic_cast<const CbmGlobalTrack*>(trackObj);
        CbmKfTrackFitter::Track trkKf;
        if (!fFitter.CreateGlobalTrack(trkKf, *track)) {
          LOG(fatal) << GetName() << "::Exec: can not create the track for the fit! ";
          break;
        }
        trkKf.fMsQp0      = 1. / 0.5;
        trkKf.fIsMsQp0Set = true;

        // minimalistic QA tool for tracks used for target projection
        int nhit[(size_t) ECbmModuleId::kLastModule] = {0};
        for (auto& n : trkKf.fNodes) {
          // check if hit is well defined
          if (n.fHitSystemId == ECbmModuleId::kNotExist) continue;
          // check if detector is registered
          if (fDetQa.find(n.fHitSystemId) == fDetQa.end()) continue;

          auto det = fDetQa[n.fHitSystemId];
          // det.Print();
          auto view = det.FindView(n.fMxy.X()[0], n.fMxy.Y()[0], n.fZ);
          if (!view) {
            LOG(debug) << GetName() << "::Exec: view for tracking layer " << ToString(det.id) << " not defined.";
            continue;
          }

          n.fIsTimeSet = false;
          n.fIsXySet   = false;
          trkKf.MakeConsistent();
          fFitter.FitTrack(trkKf);
          view->Load(&n);
          nhit[(int) n.fHitSystemId]++;
          n.fIsTimeSet = true;
          n.fIsXySet   = true;
        }

        // Fit track for intermediate projections
        int iprj(0);
        if (!nnodes) nnodes = (int) trkKf.fNodes.size();
        for (auto zprj : fProjs) {
          CbmKfTrackFitter::FitNode n = CbmKfTrackFitter::FitNode();
          n.fZ                        = zprj.first;
          n.fIsXySet                  = true;
          n.fReference1               = iprj++;
          trkKf.fNodes.push_back(n);
        }
        trkKf.MakeConsistent();
        fFitter.FitTrack(trkKf);
        for (auto& n : trkKf.fNodes) {
          if (n.fReference1 < 0) continue;
          if (fProjs.find(n.fReference1 + nnodes) == fProjs.end()) {
            LOG(debug) << GetName() << "::Exec: view for tracking layer " << n.fReference1 + nnodes << " not defined.";
            continue;
          }

          // View* mod = &fProjs[n.fReference1 + nnodes][0];
          // mod->Load(&n);
          //
          // if (mod->hdXx.second)
          //   mod->hdXx.second->Fill(mod->hdXx.first * n.fTrack.X()[0], mod->hdXx.first * n.fTrack.Y()[0]);
          // if (mod->hdYy.second) mod->hdYy.second->Fill(mod->hdYy.first * n.fTrack.X()[0], n.fTrack.Time()[0]);
        }
        itrack++;
      }  // END TRACK LOOP
      iev++;
    }  // END EVENT LOOP

    LOG(info) << GetName() << "::Exec : Evs(%)[" << 1.e2 * iev / fEvents->GetEntriesFast() << "] Trks(%)["
              << (fGTracks ? 1.e3 * itrack / fGTracks->GetEntriesFast() : 0) << "]";
    return;  // END EbyE QA
  }

  // FULL TS LOOP
  for (auto detp : fDetQa) {  // SYSTEM LOOP
    auto det = detp.second;
    // check SYSTEM @ hit level
    if (!fHits[det.id]) {
      LOG(error) << GetName() << "::Exec() : Hits for " << ToString(det.id) << " not available. Skip.";
      continue;
    }
    int nh = fHits[det.id]->GetEntriesFast();

    for (int ih(0); ih < nh; ih++) {  // HIT LOOP
      const CbmHit* hit = dynamic_cast<CbmHit*>(fHits[det.id]->At(ih));
      // printf("%s hit_%2d dt=%f time_event=%f time_hit=%f\n", ToString(det.id).data(), ih, ev->GetStartTime() - hit->GetTime(), ev->GetStartTime(), hit->GetTime());

      bool ret(false);
      for (auto view : det.fViews) {  // VIEW LOOP
        switch (det.id) {
          case ECbmModuleId::kMvd: ret = view.Load<CbmMvdHit>(hit); break;
          case ECbmModuleId::kSts: ret = view.Load<CbmStsHit>(hit); break;
          case ECbmModuleId::kMuch: ret = view.Load<CbmMuchPixelHit>(hit); break;
          case ECbmModuleId::kRich: ret = view.Load<CbmRichHit>(hit); break;
          case ECbmModuleId::kTrd: ret = view.Load<CbmTrdHit>(hit); break;
          case ECbmModuleId::kTrd2d: ret = view.Load<CbmTrdHit>(hit); break;
          case ECbmModuleId::kTof: ret = view.Load<CbmTofHit>(hit); break;
          case ECbmModuleId::kFsd: ret = view.Load<CbmFsdHit>(hit); break;
          case ECbmModuleId::kPsd: ret = view.Load<CbmPsdHit>(hit); break;
          default: LOG(fatal) << GetName() << "::Exec : unsupported det " << ToString(det.id); break;
        }
        if (ret) break;
      }  // END VIEW LOOP
    }    // END HIT LOOP
  }      // END SYSTEM LOOP
  if (!fGTracks) {
    LOG(info) << GetName() << "::Exec : TS local reco only.";
    return;  // END TS QA (no tracking)
  }

  for (auto trackObj : *fGTracks) {  // TRACK LOOP
    auto track = dynamic_cast<const CbmGlobalTrack*>(trackObj);

    // track QA filtering
    if (!FilterTrack(track)) continue;

    CbmKfTrackFitter::Track trkKf;
    if (!fFitter.CreateGlobalTrack(trkKf, *track)) {
      LOG(fatal) << GetName() << "::Exec: can not create the track for the fit! ";
      break;
    }
    trkKf.fMsQp0      = 1. / 0.5;
    trkKf.fIsMsQp0Set = true;
    if (!nnodes) nnodes = (int) trkKf.fNodes.size();

    // minimalistic QA tool for tracks used for target projection
    int nhit[(size_t) ECbmModuleId::kLastModule] = {0};
    for (auto& n : trkKf.fNodes) {
      // check if hit is well defined
      if (n.fHitSystemId == ECbmModuleId::kNotExist) continue;
      // check if detector is registered
      if (fDetQa.find(n.fHitSystemId) == fDetQa.end()) continue;

      auto det = fDetQa[n.fHitSystemId];
      // det.Print();
      auto view = det.FindView(n.fMxy.X()[0], n.fMxy.Y()[0], n.fZ);
      if (!view) {
        LOG(debug) << GetName() << "::Exec: view for tracking layer " << ToString(det.id) << " not defined.";
        continue;
      }

      n.fIsTimeSet = false;
      n.fIsXySet   = false;
      trkKf.MakeConsistent();
      fFitter.FitTrack(trkKf);
      view->Load(&n);
      nhit[(int) n.fHitSystemId]++;
      n.fIsTimeSet = true;
      n.fIsXySet   = true;
    }

    // Fit track with all hits ON for vx projection
    int iprj(0);
    for (auto zprj : fProjs) {
      CbmKfTrackFitter::FitNode n = CbmKfTrackFitter::FitNode();
      n.fZ                        = zprj.first;
      n.fIsXySet                  = true;
      n.fReference1               = iprj++;
      trkKf.fNodes.push_back(n);
    }
    trkKf.MakeConsistent();
    fFitter.FitTrack(trkKf);
    for (auto& n : trkKf.fNodes) {
      if (n.fReference1 < 0) continue;
      if (fProjs.find(n.fReference1 + nnodes) == fProjs.end()) {
        LOG(debug) << GetName() << "::Exec: view for tracking layer " << n.fReference1 + nnodes << " not defined.";
        continue;
      }

      // View* mod = &fProjs[n.fReference1 + nnodes][0];
      // mod->Load(&n);
      // if (mod->hdXx.second)
      //   mod->hdXx.second->Fill(mod->hdXx.first * n.fTrack.X()[0], mod->hdXx.first * n.fTrack.Y()[0]);
      // if (mod->hdYy.second) mod->hdYy.second->Fill(mod->hdYy.first * n.fTrack.X()[0], n.fTrack.Time()[0]);
    }
    itrack++;
  }  // END TRACK LOOP
  LOG(info) << GetName() << "::Exec : Trks(%)[" << (fGTracks ? 1.e3 * itrack / fGTracks->GetEntriesFast() : 0) << "]";

  return;  // END TS QA (tracking)
}

//_____________________________________________________________________
void CbmRecoQaTask::SetProjections(std::vector<double> vzpj)
{
  fProjs.clear();
  for (auto zprj : vzpj) {
    // fProjView[zprj] = new TH2D(Form("hxx"), Form("; X (cm); Y (cm)"), 450, -25, 20, 450, -25, 20);
    fProjs.emplace(zprj, Detector::View(Form("z%.2f", zprj), "", {}));
  }
}

//_____________________________________________________________________
void CbmRecoQaTask::Finish()
{
  FairSink* sink = FairRootManager::Instance()->GetSink();
  sink->WriteObject(&fOutFolder, nullptr);
}

//_____________________________________________________________________
bool CbmRecoQaTask::FilterEvent(const CbmEvent* ptr)
{
  for (auto evCut : fFilterEv) {
    if (!evCut.Accept(ptr, this)) return false;
  }
  return true;
}

//_____________________________________________________________________
bool CbmRecoQaTask::FilterTrack(const CbmGlobalTrack* ptr)
{
  for (auto trkCut : fFilterTrk) {
    if (!trkCut.Accept(ptr, this)) return false;
  }
  return true;
}

//_____________________________________________________________________
void CbmRecoQaTask::InitMcbm22()
{
  CbmSetup* setup = CbmSetup::Instance();
  if (!setup) {
    LOG(fatal) << GetName() << "::InitMcbm22() : Missing setup definition.";
    return;
  }

  TString dtag;
  Detector::View* v(nullptr);
  if (setup->IsActive(ECbmModuleId::kSts)) {
    dtag          = "v22f_mcbm";
    Detector* sts = AddDetector(ECbmModuleId::kSts);
    // U0 L0 M0
    v = sts->AddView(
      "U0L0M0",
      Form("/cave_1/sts_%s_0/Station01_1/Ladder09_1/HalfLadder09d_2/HalfLadder09d_Module03_1/Sensor03_1", dtag.Data()),
      {0, 0, 0});
    v->SetProjection(eViewProjection::kXdX, 1, "mm");
    v->SetProjection(eViewProjection::kYdY, 4, "mm");
    v = sts->AddView(
      "U0L0M1",
      Form("/cave_1/sts_%s_0/Station01_1/Ladder09_1/HalfLadder09d_2/HalfLadder09d_Module03_2/Sensor03_1", dtag.Data()),
      {0, 0, 1});
    v->SetProjection(eViewProjection::kXdX, 1, "mm");
    v->SetProjection(eViewProjection::kYdY, 4, "mm");
    // U0 L1
    v = sts->AddView(
      "U0L1M0",
      Form("/cave_1/sts_%s_0/Station01_1/Ladder09_2/HalfLadder09d_2/HalfLadder09d_Module03_1/Sensor03_1", dtag.Data()),
      {0, 1, 0});
    v->SetProjection(eViewProjection::kXdX, 1, "mm");
    v->SetProjection(eViewProjection::kYdY, 4, "mm");
    v = sts->AddView(
      "U0L1M1",
      Form("/cave_1/sts_%s_0/Station01_1/Ladder09_2/HalfLadder09d_2/HalfLadder09d_Module03_2/Sensor03_1", dtag.Data()),
      {0, 1, 1});
    v->SetProjection(eViewProjection::kXdX, 1, "mm");
    v->SetProjection(eViewProjection::kYdY, 4, "mm");
    // U1 L0
    v = sts->AddView(
      "U1L0M0",
      Form("/cave_1/sts_%s_0/Station02_2/Ladder10_1/HalfLadder10d_2/HalfLadder10d_Module03_1/Sensor03_1", dtag.Data()),
      {1, 0, 0});
    v->SetProjection(eViewProjection::kXdX, 0.75, "mm");
    v->SetProjection(eViewProjection::kYdY, 3, "mm");
    v = sts->AddView(
      "U1L0M1",
      Form("/cave_1/sts_%s_0/Station02_2/Ladder10_1/HalfLadder10d_2/HalfLadder10d_Module04_2/Sensor04_1", dtag.Data()),
      {1, 0, 1});
    v->SetProjection(eViewProjection::kXdX, 0.75, "mm");
    v->SetProjection(eViewProjection::kYdY, 3, "mm");
    // U1 L1
    v = sts->AddView(
      "U1L1M0",
      Form("/cave_1/sts_%s_0/Station02_2/Ladder12_2/HalfLadder12d_2/HalfLadder12d_Module03_1/Sensor03_1", dtag.Data()),
      {1, 1, 0});
    v->SetProjection(eViewProjection::kXdX, 0.75, "mm");
    v->SetProjection(eViewProjection::kYdY, 3, "mm");
    v = sts->AddView(
      "U1L1M1",
      Form("/cave_1/sts_%s_0/Station02_2/Ladder12_2/HalfLadder12d_2/HalfLadder12d_Module04_2/Sensor04_1", dtag.Data()),
      {1, 1, 1});
    v->SetProjection(eViewProjection::kXdX, 0.75, "mm");
    v->SetProjection(eViewProjection::kYdY, 3, "mm");
    // U1 L2
    v = sts->AddView(
      "U1L2M0",
      Form("/cave_1/sts_%s_0/Station02_2/Ladder11_3/HalfLadder11d_2/HalfLadder11d_Module03_1/Sensor03_1", dtag.Data()),
      {1, 2, 0});
    v->SetProjection(eViewProjection::kXdX, 0.75, "mm");
    v->SetProjection(eViewProjection::kYdY, 3, "mm");
    v = sts->AddView(
      "U1L2M1",
      Form("/cave_1/sts_%s_0/Station02_2/Ladder11_3/HalfLadder11d_2/HalfLadder11d_Module03_2/Sensor03_1", dtag.Data()),
      {1, 2, 1});
    v->SetProjection(eViewProjection::kXdX, 0.75, "mm");
    v->SetProjection(eViewProjection::kYdY, 3, "mm");
    v = sts->AddView(
      "U1L2M2",
      Form("/cave_1/sts_%s_0/Station02_2/Ladder11_3/HalfLadder11d_2/HalfLadder11d_Module03_3/Sensor03_1", dtag.Data()),
      {1, 2, 2});
    v->SetProjection(eViewProjection::kXdX, 0.75, "mm");
    v->SetProjection(eViewProjection::kYdY, 3, "mm");
  }
  if (setup->IsActive(ECbmModuleId::kTrd)) {
    dtag          = "v22h_mcbm";
    Detector* trd = AddDetector(ECbmModuleId::kTrd2d);
    // Trd2D
    v = trd->AddView("2D", Form("/cave_1/trd_%s_0/layer01_20101/module9_101001001/gas_0", dtag.Data()), {5});
    v->SetProjection(eViewProjection::kChdT, 400, "ns");
    v->SetProjection(eViewProjection::kXdX, 10, "mm");
    v->SetProjection(eViewProjection::kYdY, 20, "mm");
    // Trd1Dx
    trd = AddDetector(ECbmModuleId::kTrd);
    v   = trd->AddView("1Dx", Form("/cave_1/trd_%s_0/layer02_10202/module8_101002001", dtag.Data()), {21});
    v->SetProjection(eViewProjection::kChdT, 400, "ns");
    v->SetProjection(eViewProjection::kXdX, 1.5, "cm");
    // Trd1Dy
    v = trd->AddView("1Dy", Form("/cave_1/trd_%s_0/layer03_11303/module8_101303001", dtag.Data()), {37});
    v->SetProjection(eViewProjection::kChdT, 400, "ns");
  }
  if (setup->IsActive(ECbmModuleId::kTof)) {
    dtag          = "v21k_mcbm";
    Detector* tof = AddDetector(ECbmModuleId::kTof);
    tof->hit.name = "TofCalHit";
    vector<int> tofSelect(3);
    // add type 0
    tofSelect[1] = 0;  // ToF type
    for (int ism(0); ism < 4; ism++) {
      tofSelect[0] = ism;
      for (int irpc(0); irpc < 5; irpc++) {
        tofSelect[2] = irpc;
        v            = tof->AddView(Form("Sm%dRpc%d", tofSelect[0], tofSelect[2]),
                         Form("/cave_1/tof_%s_0/tof_%sStand_1/module_%d_%d/gas_box_0/counter_%d", dtag.Data(),
                              dtag.Data(), tofSelect[1], tofSelect[0], tofSelect[2]),
                         tofSelect);
        v->SetProjection(eViewProjection::kChdT, 15, "ns");
      }
    }
    // add type 2
    tofSelect[1] = 2;  // ToF type
    tofSelect[0] = 0;  // Tof SM id
    for (int irpc(0); irpc < 5; irpc++) {
      tofSelect[2] = irpc;
      tof->AddView(Form("Sm_2%dRpc%d", tofSelect[0], tofSelect[2]),
                   Form("/cave_1/tof_%s_0/tof_%sStand_1/module_%d_%d/gas_box_0/counter_%d", dtag.Data(), dtag.Data(),
                        tofSelect[1], tofSelect[0], tofSelect[2]),
                   tofSelect);
    }
  }

  // ===============================================================================
  // TRG - upstream projections
  // int iprjLy = 0;
  // for (auto prj : fProjs) {
  //   LOG(debug) << GetName() << "::Init: Project trks @ " << prj.first;
  //   prj.second.name       = Form("Z=%+4.1f", prj.first);
  //   prj.second.hdXx.first = 1;
  //   prj.second.hdXx.second =
  //     new TH2D(Form("Hxy%d", iprjLy), Form("Project @ z = %+5.2f; X_{trk} (cm); Y_{trk} (cm)", prj.first),
  //               700, -25, 10, 100, -5, 5);
  //   prj.second.hdT.first = 1;
  //   prj.second.hdT.second =
  //     new TH2D(Form("Hxt%d", iprjLy), Form("Project @ z = %+5.2f; X_{trk} (cm); T_{trk} - T_{ev} (ns)", prj.first),
  //               700, -25, 10, 350, -15.5, 5.5);
  //   iprjLy++;
  // }
}

//_____________________________________________________________________
void CbmRecoQaTask::InitMcbm24()
{
  CbmSetup* setup = CbmSetup::Instance();
  if (!setup) {
    LOG(fatal) << GetName() << "::InitMcbm24() : Missing setup definition.";
    return;
  }
  TString dtag;
  Detector::View* v(nullptr);
  if (setup->IsActive(ECbmModuleId::kSts)) {
    // setup->GetGeoTag(ECbmModuleId::kSts, dtag);
    // LOG(debug) << GetName() << "::InitMcbm24() : found " << dtag;
    dtag          = "v24c_mcbm";
    Detector* sts = AddDetector(ECbmModuleId::kSts);
    // U0 L0 M0
    v = sts->AddView(
      "U0L0M0",
      Form("/cave_1/sts_%s_0/Station01_1/Ladder13_1/HalfLadder13u_1/HalfLadder13u_Module03_1/Sensor03_1", dtag.Data()),
      {0, 0, 0});
    v->SetProjection(eViewProjection::kYdY, 5, "mm");
    v->SetProjection(eViewProjection::kXdX, 1, "mm");
    // U1 L0 M0
    v = sts->AddView(
      "U1L0M0",
      Form("/cave_1/sts_%s_0/Station02_2/Ladder09_1/HalfLadder09d_2/HalfLadder09d_Module03_1/Sensor03_1", dtag.Data()),
      {1, 0, 0});
    v->SetProjection(eViewProjection::kYdY, 5, "mm");
    v->SetProjection(eViewProjection::kXdX, 1, "mm");
    v = sts->AddView(
      "U1L0M1",
      Form("/cave_1/sts_%s_0/Station02_2/Ladder09_1/HalfLadder09d_2/HalfLadder09d_Module03_2/Sensor03_1", dtag.Data()),
      {1, 0, 1});
    v->SetProjection(eViewProjection::kYdY, 5, "mm");
    v->SetProjection(eViewProjection::kXdX, 1, "mm");
    // U1 L1
    v = sts->AddView(
      "U1L1M0",
      Form("/cave_1/sts_%s_0/Station02_2/Ladder09_2/HalfLadder09d_2/HalfLadder09d_Module03_1/Sensor03_1", dtag.Data()),
      {1, 1, 0});
    v->SetProjection(eViewProjection::kYdY, 5, "mm");
    v->SetProjection(eViewProjection::kXdX, 1, "mm");
    v = sts->AddView(
      "U1L1M1",
      Form("/cave_1/sts_%s_0/Station02_2/Ladder09_2/HalfLadder09d_2/HalfLadder09d_Module03_2/Sensor03_1", dtag.Data()),
      {1, 1, 1});
    v->SetProjection(eViewProjection::kYdY, 5, "mm");
    v->SetProjection(eViewProjection::kXdX, 1, "mm");
    // U2 L0
    v = sts->AddView(
      "U2L0M0",
      Form("/cave_1/sts_%s_0/Station03_3/Ladder10_1/HalfLadder10d_2/HalfLadder10d_Module03_1/Sensor03_1", dtag.Data()),
      {2, 0, 0});
    v->SetProjection(eViewProjection::kYdY, 5, "mm");
    v->SetProjection(eViewProjection::kXdX, 1, "mm");
    v = sts->AddView(
      "U2L0M1",
      Form("/cave_1/sts_%s_0/Station03_3/Ladder10_1/HalfLadder10d_2/HalfLadder10d_Module04_2/Sensor04_1", dtag.Data()),
      {2, 0, 1});
    v->SetProjection(eViewProjection::kYdY, 5, "mm");
    v->SetProjection(eViewProjection::kXdX, 1, "mm");
    // U2 L1
    v = sts->AddView(
      "U2L1M0",
      Form("/cave_1/sts_%s_0/Station03_3/Ladder12_2/HalfLadder12d_2/HalfLadder12d_Module03_1/Sensor03_1", dtag.Data()),
      {2, 1, 0});
    v->SetProjection(eViewProjection::kYdY, 5, "mm");
    v->SetProjection(eViewProjection::kXdX, 1, "mm");
    v = sts->AddView(
      "U2L1M1",
      Form("/cave_1/sts_%s_0/Station03_3/Ladder12_2/HalfLadder12d_2/HalfLadder12d_Module04_2/Sensor04_1", dtag.Data()),
      {2, 1, 1});
    v->SetProjection(eViewProjection::kYdY, 5, "mm");
    v->SetProjection(eViewProjection::kXdX, 1, "mm");
    // U2 L2
    v = sts->AddView(
      "U2L2M0",
      Form("/cave_1/sts_%s_0/Station03_3/Ladder11_3/HalfLadder11d_2/HalfLadder11d_Module03_1/Sensor03_1", dtag.Data()),
      {2, 2, 0});
    v->SetProjection(eViewProjection::kYdY, 5, "mm");
    v->SetProjection(eViewProjection::kXdX, 1, "mm");
    v = sts->AddView(
      "U2L2M1",
      Form("/cave_1/sts_%s_0/Station03_3/Ladder11_3/HalfLadder11d_2/HalfLadder11d_Module03_2/Sensor03_1", dtag.Data()),
      {2, 2, 1});
    v = sts->AddView(
      "U2L2M2",
      Form("/cave_1/sts_%s_0/Station03_3/Ladder11_3/HalfLadder11d_2/HalfLadder11d_Module03_3/Sensor03_1", dtag.Data()),
      {2, 2, 2});
    v->SetProjection(eViewProjection::kYdY, 5, "mm");
    v->SetProjection(eViewProjection::kXdX, 1, "mm");
  }
  if (setup->IsActive(ECbmModuleId::kTrd)) {
    // setup->GetGeoTag(ECbmModuleId::kTrd, dtag);
    // LOG(debug) << GetName() << "::InitMcbm24() : found " << dtag;
    dtag          = "v24e_mcbm";
    Detector* trd = AddDetector(ECbmModuleId::kTrd2d);
    // Trd2D
    v = trd->AddView("2D", Form("/cave_1/trd_%s_0/layer01_20101/module9_101001001/gas_0", dtag.Data()), {5});
    v->SetProjection(eViewProjection::kChdT, 300, "ns");
    // Trd1Dx
    trd = AddDetector(ECbmModuleId::kTrd);
    v   = trd->AddView("1Dx", Form("/cave_1/trd_%s_0/layer02_10202/module5_101002001", dtag.Data()), {21});
    v->SetProjection(eViewProjection::kChdT, 300, "ns");
    v->SetProjection(eViewProjection::kXdX, 7, "cm");
    // Trd1Dy
    v = trd->AddView("1Dy", Form("/cave_1/trd_%s_0/layer03_11303/module5_101103001", dtag.Data()), {37});
    v->SetProjection(eViewProjection::kChdT, 300, "ns");
    v->SetProjection(eViewProjection::kXdX, 7, "cm");
  }
  if (setup->IsActive(ECbmModuleId::kTof)) {
    // setup->GetGeoTag(ECbmModuleId::kTof, dtag);
    // LOG(debug) << GetName() << "::InitMcbm24() : found " << dtag;
    dtag          = "v24d_mcbm";
    Detector* tof = AddDetector(ECbmModuleId::kTof);
    vector<int> tofSelect(3);
    // add type 0
    for (int ism(0); ism < 6; ism++) {
      tofSelect[0] = ism;
      for (int irpc(0); irpc < 5; irpc++) {
        tofSelect[2] = irpc;
        v            = tof->AddView(Form("Sm%dRpc%d", ism, irpc),
                         Form("/cave_1/tof_%s_0/tof_%sStand_1/module_%d_%d/gas_box_0/counter_%d", dtag.Data(),
                              dtag.Data(), tofSelect[1], tofSelect[0], tofSelect[2]),
                         tofSelect);
        v->SetProjection(eViewProjection::kXdX, 7, "cm");
        v->SetProjection(eViewProjection::kYdY, 7, "cm");
      }
    }
    // add type 6 (Buch)
    tofSelect[0] = 0;
    tofSelect[1] = 6;
    for (int irpc(0); irpc < 2; irpc++) {
      tofSelect[2] = irpc;
      tof->AddView(Form("BuchRpc%d", irpc),
                   Form("/cave_1/tof_%s_0/tof_%sStand_1/module_%d_%d/gas_box_0/counter_%d", dtag.Data(), dtag.Data(),
                        tofSelect[1], tofSelect[0], tofSelect[2]),
                   tofSelect);
    }
    // add type 9
    tofSelect[1] = 9;
    for (int ism(0); ism < 2; ism++) {
      tofSelect[0] = ism;
      for (int irpc(0); irpc < 2; irpc++) {
        tofSelect[2] = irpc;
        tof->AddView(Form("Test%dRpc%d", ism, irpc),
                     Form("/cave_1/tof_%s_0/tof_%sStand_1/module_%d_%d/gas_box_0/counter_%d", dtag.Data(), dtag.Data(),
                          tofSelect[1], tofSelect[0], tofSelect[2]),
                     tofSelect);
      }
    }
    // add type 2
    tofSelect[1] = 2;
    for (int ism(0); ism < 2; ism++) {
      tofSelect[0] = ism;
      for (int irpc(0); irpc < 5; irpc++) {
        tofSelect[2] = irpc;
        tof->AddView(Form("Sm2%dRpc%d", ism, irpc),
                     Form("/cave_1/tof_%s_0/tof_%sStand_1/module_%d_%d/gas_box_0/counter_%d", dtag.Data(), dtag.Data(),
                          tofSelect[1], tofSelect[0], tofSelect[2]),
                     tofSelect);
      }
    }
  }
  if (setup->IsActive(ECbmModuleId::kRich)) {
    // setup->GetGeoTag(ECbmModuleId::kRich, dtag);
    // LOG(debug) << GetName() << "::InitMcbm24() : found " << dtag;
    dtag           = "v24a_mcbm";
    Detector* rich = AddDetector(ECbmModuleId::kRich);
    // Aerogel 1
    rich->AddView("Aerogel", Form("/cave_1/rich_%s_0/box_1/Gas_1", dtag.Data()), {4});
  }
}

//========= DETECTOR ======================
CbmRecoQaTask::Detector::Detector(ECbmModuleId did)
{
  id = did;
  switch (id) {
    case ECbmModuleId::kMvd: new (&hit) Data(ECbmDataType::kMvdHit, "MvdHit"); break;
    case ECbmModuleId::kSts:
      new (&hit) Data(ECbmDataType::kStsHit, "StsHit");
      new (&trk) Data(ECbmDataType::kStsTrack, "StsTrack");
      break;
    case ECbmModuleId::kMuch: new (&hit) Data(ECbmDataType::kMuchPixelHit, "MuchHit"); break;
    case ECbmModuleId::kRich: new (&hit) Data(ECbmDataType::kRichHit, "RichHit"); break;
    case ECbmModuleId::kTrd:
    case ECbmModuleId::kTrd2d:
      new (&hit) Data(ECbmDataType::kTrdHit, "TrdHit");
      new (&trk) Data(ECbmDataType::kTrdTrack, "TrdTrack");
      break;
    case ECbmModuleId::kTof:
      new (&hit) Data(ECbmDataType::kTofHit, "TofHit");
      new (&trk) Data(ECbmDataType::kTofTrack, "TofTrack");
      break;
    case ECbmModuleId::kFsd: new (&hit) Data(ECbmDataType::kFsdHit, "FsdHit"); break;
    case ECbmModuleId::kPsd: new (&hit) Data(ECbmDataType::kPsdHit, "PsdHit"); break;
    default:
      LOG(warn) << "QA unsupported for Detector=" << ToString(did);
      id = ECbmModuleId::kNotExist;
      break;
  }
}

//+++++++++++++++++++++  Detector  ++++++++++++++++++++++++++++++++
//_________________________________________________________________
CbmRecoQaTask::Detector::View* CbmRecoQaTask::Detector::AddView(const char* n, const char* p, std::vector<int> set)
{
  View* v(nullptr);
  fViews.emplace_back(n, p, set);
  v = &fViews.back();
  v->AddProjection(CbmRecoQaTask::eViewProjection::kXYs);

  v->AddProjection(CbmRecoQaTask::eViewProjection::kChdT);
  v->AddProjection(CbmRecoQaTask::eViewProjection::kXYh);

  v->AddProjection(CbmRecoQaTask::eViewProjection::kXYa);
  v->AddProjection(CbmRecoQaTask::eViewProjection::kXYp);
  v->AddProjection(CbmRecoQaTask::eViewProjection::kXdX);
  v->AddProjection(CbmRecoQaTask::eViewProjection::kYdY);
  v->AddProjection(CbmRecoQaTask::eViewProjection::kWdT);
  v->AddProjection(CbmRecoQaTask::eViewProjection::kXpX);
  v->AddProjection(CbmRecoQaTask::eViewProjection::kYpY);

  return v;
}

CbmRecoQaTask::Detector::View* CbmRecoQaTask::Detector::GetView(const char* n)
{
  for (auto& view : fViews)
    if (view.name.compare(n) == 0) return &view;

  return nullptr;
}

CbmRecoQaTask::Detector::View* CbmRecoQaTask::Detector::FindView(double x, double y, double z)
{
  for (auto& v : fViews) {
    // printf("Looking for p[%f %f %f] in V[%s-%s] @ %f %f %f (%f %f %f)\n", x, y, z, ToString(id).data(), v.name.data(), v.pos[0], v.pos[1], v.pos[2], v.size[0], v.size[1], v.size[2]);
    if (abs(v.pos[0] - x) > 0.5 * v.size[0]) continue;
    if (abs(v.pos[1] - y) > 0.5 * v.size[1]) continue;
    if (abs(v.pos[2] - z) > 0.5 * v.size[2]) continue;
    return &v;
  }

  return nullptr;
}

bool CbmRecoQaTask::Detector::Init(TDirectoryFile* fOut)
{
  /* Query the geometry and trigger detector views init
  */
  if (!gGeoManager) {
    LOG(fatal) << "CbmRecoQaTask::Detector::Init() " << ToString(id) << " missing geometry.";
    return false;
  }
  LOG(debug) << "CbmRecoQaTask::Detector::Init() : " << ToString(id);

  TDirectoryFile* modDir =
    (TDirectoryFile*) fOut->mkdir(ToString(id).data(), Form("Reco QA for %s", ToString(id).data()));
  bool ret(true);
  for (auto& view : fViews) {
    bool vret = view.Init(ToString(id).data());
    view.Register(modDir);
    ret = ret && vret;
  }
  return ret;
}

void CbmRecoQaTask::Detector::Print() const
{
  stringstream s;
  s << "D[" << ToString(id) << "] views[" << fViews.size() << "]\n";
  for (auto v : fViews)
    s << v.ToString();
  cout << s.str();
}

//+++++++++++++++++++++  Detector::View  ++++++++++++++++++++++++++++++++
//_______________________________________________________________________
bool CbmRecoQaTask::Detector::View::AddProjection(eViewProjection prj, float range, const char* unit)
{
  if (fProjection.find(prj) != fProjection.end()) {
    LOG(warn) << "Projection " << ToString(prj) << " already registered";
    return false;
  }
  int scale(1);
  if (strcmp(unit, "mm") == 0)
    scale = 10;
  else if (strcmp(unit, "um") == 0)
    scale = 10000;
  else if (strcmp(unit, "ps") == 0)
    scale = 1000;
  else if (strcmp(unit, "eV") == 0)
    scale = 1000;
  else if (strcmp(unit, "cm") == 0)
    scale = 1;
  else if (strcmp(unit, "ns") == 0)
    scale = 1;
  else if (strcmp(unit, "keV") == 0)
    scale = 1;
  else if (strcmp(unit, "a.u.") == 0)
    scale = 1;
  else
    LOG(warn) << "Projection units " << unit << " not registered. Natural units will be used.";

  fProjection[prj] = make_tuple(scale, range, nullptr);
  return true;
}

bool CbmRecoQaTask::Detector::View::SetProjection(eViewProjection prj, float range, const char* unit)
{
  if (fProjection.find(prj) == fProjection.end()) {
    LOG(warn) << "Projection " << ToString(prj)
              << " not initialized. Calling \"CbmRecoQaTask::Detector::View::AddProjection()\"";
    return AddProjection(prj, range, unit);
  }
  int scale(1);
  if (strcmp(unit, "mm") == 0)
    scale = 10;
  else if (strcmp(unit, "um") == 0)
    scale = 10000;
  else if (strcmp(unit, "ps") == 0)
    scale = 1000;
  else if (strcmp(unit, "eV") == 0)
    scale = 1000;
  else if (strcmp(unit, "cm") == 0)
    scale = 1;
  else if (strcmp(unit, "ns") == 0)
    scale = 1;
  else if (strcmp(unit, "keV") == 0)
    scale = 1;
  else if (strcmp(unit, "a.u.") == 0)
    scale = 1;
  else
    LOG(warn) << "Projection units " << unit << " not registered. Natural units will be used.";

  get<0>(fProjection[prj]) = scale;
  get<1>(fProjection[prj]) = range;
  return true;
}

bool CbmRecoQaTask::Detector::View::Init(const char* dname)
{
  if (!gGeoManager->cd(path.data())) return false;
  TGeoHMatrix* m = gGeoManager->GetCurrentMatrix();
  const double* tr(m->GetTranslation());
  TGeoVolume* v = gGeoManager->GetCurrentVolume();
  TGeoShape* bb = v->GetShape();
  double w_lo, w_hi;
  for (int i(0); i < 3; i++) {  // loop over the spatial dimensions
    size[i] = bb->GetAxisRange(i + 1, w_lo, w_hi);
    pos[i]  = 0.5 * (w_lo + w_hi) + tr[i];
  }

  LOG(info) << "CbmRecoQaTask::Detector(" << dname << ")::View(" << name << ")::Init() : size [" << size[0] << " x "
            << size[1] << " x " << size[2] << "].";

  // TODO short-hand for XY display resolution
  int dscale = 10;
  if (strcmp(dname, "Tof") == 0 || strcmp(dname, "Rich") == 0) dscale = 1;

  int nbinsx, nbinsy;
  string unit;
  for (auto& projection : fProjection) {
    int scale    = get<0>(projection.second);
    float yrange = get<1>(projection.second);
    char xy_id   = 0;
    double xlo = pos[0] - 0.5 * size[0], xhi = pos[0] + 0.5 * size[0], ylo = pos[1] - 0.5 * size[1],
           yhi = pos[1] + 0.5 * size[1];
    switch (projection.first) {
      case eViewProjection::kXdX:
        nbinsx                    = 10 * ceil(size[0]);  // mm binning
        unit                      = makeYrange(scale, yrange);
        nbinsy                    = 200;  //* ceil(yrange);
        get<2>(projection.second) = new TH2D(Form("hxx_%s%s", dname, name.data()),
                                             Form("X resolution %s [%s]; X_{%s-%s} (cm); #Delta X (%s)", name.data(),
                                                  dname, dname, name.data(), unit.data()),
                                             nbinsx, xlo, xhi, nbinsy, -yrange, yrange);
        break;

      case eViewProjection::kYdY:
        nbinsx                    = 10 * ceil(size[1]);  // mm binning
        unit                      = makeYrange(scale, yrange);
        nbinsy                    = 200;  //* ceil(yrange);
        get<2>(projection.second) = new TH2D(Form("hyy_%s%s", dname, name.data()),
                                             Form("Y resolution %s [%s]; Y_{%s-%s} (cm); #Delta Y (%s)", name.data(),
                                                  dname, dname, name.data(), unit.data()),
                                             nbinsx, ylo, yhi, nbinsy, -yrange, yrange);
        break;

      case eViewProjection::kXpX:
        nbinsx = 10 * ceil(size[0]);  // mm binning
        if (yrange < 0) {
          LOG(debug) << "ProjectionP[" << name << "] using default range.";
          yrange = 5;  // +- 5 sigma range
        }
        nbinsy = 200;  //* ceil(yrange);
        get<2>(projection.second) =
          new TH2D(Form("hpx_%s%s", dname, name.data()),
                   Form("X pulls %s [%s]; X_{%s-%s} (cm); pull(X)", name.data(), dname, dname, name.data()), nbinsx,
                   xlo, xhi, nbinsy, -yrange, yrange);
        break;

      case eViewProjection::kYpY:
        nbinsx = 10 * ceil(size[1]);  // mm binning
        if (yrange < 0) {
          LOG(debug) << "ProjectionP[" << name << "] using default range.";
          yrange = 5;  // +- 5 sigma range
        }
        nbinsy = 200;  //* ceil(yrange);
        get<2>(projection.second) =
          new TH2D(Form("hpy_%s%s", dname, name.data()),
                   Form("Y pulls %s [%s]; Y_{%s-%s} (cm); pull(Y)", name.data(), dname, dname, name.data()), nbinsx,
                   ylo, yhi, nbinsy, -yrange, yrange);
        break;

      case eViewProjection::kWdT:
        nbinsx                    = 10 * ceil(size[0]);  // mm binning
        unit                      = makeTrange(scale, yrange);
        nbinsy                    = 2 * ceil(yrange);
        get<2>(projection.second) = new TH2D(Form("hxt_%s%s", dname, name.data()),
                                             Form("Hit_Trk %s [%s]; X_{%s-%s} (cm); #Delta time_{TRK} (%s)",
                                                  name.data(), dname, dname, name.data(), unit.data()),
                                             nbinsx, xlo, xhi, nbinsy, -yrange, yrange);
        break;

      case eViewProjection::kChdT:
        nbinsx                    = dscale * ceil(size[0]);  // mm binning
        unit                      = makeTrange(scale, yrange);
        nbinsy                    = 10 * ceil(yrange);
        get<2>(projection.second) = new TH2D(Form("hct_%s%s", dname, name.data()),
                                             Form("Hit_Event %s [%s]; X_{%s-%s} (cm); #Delta time_{EV} (%s)",
                                                  name.data(), dname, dname, name.data(), unit.data()),
                                             nbinsx, xlo, xhi, nbinsy, -yrange, yrange);
        break;

      case eViewProjection::kXYa:
        xy_id = 'a';
        // fall through
      case eViewProjection::kXYs:
        if (!xy_id) xy_id = 's';
        // fall through
      case eViewProjection::kXYh:
        if (!xy_id) xy_id = 'h';
        nbinsx = dscale * ceil(size[0]);  // mm binning
        nbinsy = dscale * ceil(size[1]);
        get<2>(projection.second) =
          new TH2D(Form("hxy%c_%s%s", xy_id, dname, name.data()),
                   Form("Hit_{%s} %s [%s]; X_{%s-%s} (cm);  Y_{%s-%s} (cm)", (xy_id == 'h' ? "reco" : "attach"),
                        name.data(), dname, dname, name.data(), dname, name.data()),
                   nbinsx, xlo, xhi, nbinsy, ylo, yhi);
        break;
      case eViewProjection::kXYp:
        nbinsx                    = dscale * ceil(size[0] * 2.);  // mm binning
        nbinsy                    = dscale * ceil(size[1] * 2.);
        xlo                       = pos[0] - size[0];
        xhi                       = pos[0] + size[0];
        ylo                       = pos[1] - size[1];
        yhi                       = pos[1] + size[1];
        get<2>(projection.second) = new TH2D(Form("hxyp_%s%s", dname, name.data()),
                                             Form("Trk_{proj.} %s [%s]; X_{%s-%s} (cm);  Y_{%s-%s} (cm)", name.data(),
                                                  dname, dname, name.data(), dname, name.data()),
                                             nbinsx, xlo, xhi, nbinsy, ylo, yhi);
        break;
    }
  }
  return true;
}

string CbmRecoQaTask::Detector::View::ToString() const
{
  stringstream s;
  s << "V[" << name << "] path[" << path << "] @ [" << pos[0] << ", " << pos[1] << ", " << pos[2] << "] size["
    << size[0] << ", " << size[1] << ", " << size[2] << "] projections[" << fProjection.size() << "] sel[";
  for (auto ii : fSelector)
    s << ii << " ";
  s << "]\n";
  return s.str();
}

uint CbmRecoQaTask::Detector::View::Register(TDirectoryFile* fOut)
{
  uint n(0);
  TDirectoryFile* lDir = (TDirectoryFile*) fOut->mkdir(name.data(), Form("QA for vol[%s]", path.data()));
  TDirectoryFile* sDir =
    (!CbmRecoQaTask::fuRecoConfig[kRecoEvents] ? (TDirectoryFile*) lDir->mkdir("TS", "TimeSlice QA") : nullptr);
  TDirectoryFile* eDir =
    (CbmRecoQaTask::fuRecoConfig[kRecoEvents] ? (TDirectoryFile*) lDir->mkdir("event", "Local Reco QA") : nullptr);
  TDirectoryFile* tDir =
    (CbmRecoQaTask::fuRecoConfig[kRecoTracks] ? (TDirectoryFile*) lDir->mkdir("track", "CA QA") : nullptr);
  LOG(debug) << "CbmRecoQaTask::Detector::View[" << name << "]::Register() : "
             << "\n\t level \"TS \" [" << (sDir ? "ON" : "OFF") << "]"
             << "\n\t level \"EVT\" [" << (eDir ? "ON" : "OFF") << "]"
             << "\n\t level \"TRK\" [" << (tDir ? "ON" : "OFF") << "]";

  for (auto& projection : fProjection) {
    TH2* hh = get<2>(projection.second);
    if (!hh) continue;
    switch (projection.first) {
      case eViewProjection::kXYa:
      case eViewProjection::kXYp:
      case eViewProjection::kXdX:
      case eViewProjection::kYdY:
      case eViewProjection::kWdT:
      case eViewProjection::kXpX:
      case eViewProjection::kYpY:
        if (tDir) tDir->Add(hh);
        break;
      case eViewProjection::kChdT:
      case eViewProjection::kXYh:
        if (eDir) eDir->Add(hh);
        break;
      case eViewProjection::kXYs:
        if (sDir) sDir->Add(hh);
        break;
    }
    n++;
  }
  LOG(debug) << "CbmRecoQaTask::Detector::View[" << name << "]::Register() : " << n << " projections for "
             << fOut->GetName();
  return n;
}

string CbmRecoQaTask::Detector::View::ToString(eViewProjection prj)
{
  string s;
  switch (prj) {
    case eViewProjection::kXYa: s = "x-y (attach)"; break;
    case eViewProjection::kXYp: s = "x-y (track)"; break;
    case eViewProjection::kXdX: s = "dx-x"; break;
    case eViewProjection::kYdY: s = "dy-y"; break;
    case eViewProjection::kXpX: s = "pull(x)-x"; break;
    case eViewProjection::kYpY: s = "pull(y)-y"; break;
    case eViewProjection::kWdT: s = "dt(trk)-w"; break;
    case eViewProjection::kChdT: s = "dt(ev)-w"; break;
    case eViewProjection::kXYh: s = "x-y (hit)"; break;
    case eViewProjection::kXYs: s = "x-y (TS)"; break;
    default: LOG(error) << "Detector::View::ToString() : Unknown projection " << int(prj); break;
  }
  return s;
}

string CbmRecoQaTask::Detector::View::makeYrange(const int scale, float& yrange)
{
  bool kDefaultRange = false;
  string unit        = "cm";
  if (yrange < 0) {
    LOG(debug) << "ProjectionY[" << name << "] using default range.";
    kDefaultRange = true;
    yrange        = 3;  // +- 6 cm default range
  }
  if (scale == 10) {  // mm resolution
    unit = "mm";
    if (kDefaultRange) yrange = 10;  // +- 20 mm default range
  }
  else if (scale == 10000) {  // micron resolution
    unit = "#mu m";
    if (kDefaultRange) yrange = 150;  // +- 300 um default range
  }
  return unit;
}

string CbmRecoQaTask::Detector::View::makeTrange(const int scale, float& yrange)
{
  bool kDefaultRange = false;
  string unit        = "ns";
  if (yrange < 0) {
    LOG(debug) << "ProjectionT[" << name << "] using default range.";
    kDefaultRange = true;
    yrange        = 40;  // +- 80 ns default range
  }
  if (scale == 1000) {  // ps resolution
    unit = "ps";
    if (kDefaultRange) yrange = 300;  // +- 600 ps default range
  }
  return unit;
}

CbmRecoQaTask::EventFilter* CbmRecoQaTask::AddEventFilter(CbmRecoQaTask::EventFilter::eEventCut type)
{
  for (auto cut : fFilterEv) {
    if (cut.fType == type) {
      LOG(warning) << GetName() << "::AddEventFilter event filter : " << cut.ToString() << " already on the list.";
      return nullptr;
    }
  }
  fFilterEv.emplace_back(type);
  return &fFilterEv.back();
}

CbmRecoQaTask::TrackFilter* CbmRecoQaTask::AddTrackFilter(CbmRecoQaTask::TrackFilter::eTrackCut type)
{
  for (auto cut : fFilterTrk) {
    if (cut.fType == type) {
      LOG(warning) << GetName() << "::AddTrackFilter track filter : " << cut.ToString() << " already on the list.";
      return nullptr;
    }
  }
  fFilterTrk.emplace_back(type);
  return &fFilterTrk.back();
}

//+++++++++++++++++++++  EventFilter  ++++++++++++++++++++++++++++++++
//____________________________________________________________________
bool CbmRecoQaTask::EventFilter::Accept(const CbmEvent* ptr, const CbmRecoQaTask* /* lnk*/)
{
  stringstream ss;
  bool ret(true);
  size_t val(0);
  switch (fType) {
    case eEventCut::kMultTrk:
      val = ptr->GetNofData(ECbmDataType::kGlobalTrack);
      if (fMinTrack > 0 && val < size_t(fMinTrack)) {
        ss << "NofTrack[" << val << "] < min[" << fMinTrack << "].";
        ret = false;
        break;
      }
      if (ret && fMaxTrack > 0 && val > size_t(fMaxTrack)) {
        ss << "NofTrack[" << val << "] > max[" << fMaxTrack << "].";
        ret = false;
      }
      break;
    case eEventCut::kMultHit:
      val = ptr->GetNofData(ECbmDataType::kStsHit);
      if (fMultHit[0] > 0 && val > size_t(fMultHit[0])) {
        ss << "Sts hits [" << val << "] > max[" << fMultHit[0] << "].";
        ret = false;
        break;
      }
      val = ptr->GetNofData(ECbmDataType::kTrdHit);
      if (fMultHit[1] > 0 && val > size_t(fMultHit[1])) {
        ss << "Trd hits [" << val << "] > max[" << fMultHit[1] << "].";
        ret = false;
        break;
      }
      val = ptr->GetNofData(ECbmDataType::kTofHit);
      if (fMultHit[2] > 0 && val > size_t(fMultHit[2])) {
        ss << "Tof hits [" << val << "] > max[" << fMultHit[2] << "].";
        ret = false;
      }
      break;
    default: break;
  }
  if (!ret) LOG(debug2) << "Event reject for : " << ss.str();
  return ret;
}
bool CbmRecoQaTask::EventFilter::SetFilter(std::vector<float> cuts)
{
  switch (fType) {
    case eEventCut::kMultTrk:
      if (cuts.size() < 2 || cuts[1] < cuts[0]) {
        LOG(warning) << "Improper definition for event filter :\n\t" << ToString() << endl;
        HelpMess();
        return false;
      }
      fMinTrack = int(cuts[0]);
      fMaxTrack = int(cuts[1]);
      break;
    case eEventCut::kMultHit:
      if (cuts.size() < 3) {
        LOG(warning) << "Improper definition for event filter :\n\t" << ToString() << endl;
        HelpMess();
        return false;
      }
      for (int i(0); i < int(CbmRecoQaTask::EventFilter::eEventDef::kNofDetHit); i++)
        fMultHit[i] = int(cuts[i]);
      break;
    case eEventCut::kTrigger: break;
    case eEventCut::kVertex: break;
    default: break;
  }
  return true;
}
std::string CbmRecoQaTask::EventFilter::ToString() const
{
  stringstream ss;
  switch (fType) {
    case eEventCut::kMultTrk: ss << "kMultTrk : \"cut on track multiplicity"; break;
    case eEventCut::kMultHit: ss << "kMultHit : \"cut on hit multiplicity"; break;
    case eEventCut::kTrigger: ss << "kTrigger : \"cut on trigger conditions"; break;
    case eEventCut::kVertex: ss << "kVertex : \"cut on vertex definition"; break;
    default: break;
  }
  return ss.str();
}
void CbmRecoQaTask::EventFilter::HelpMess() const
{
  LOG(info) << "CbmRecoQaTask::EventFilter : Usage";
  switch (fType) {
    case eEventCut::kMultTrk:
      LOG(info) << ToString();
      LOG(info) << "\tDepends one two values : min and max of the no of global tracks / event.";
      LOG(info) << "\tValue should follow the relation max >= min.";
      break;
    case eEventCut::kMultHit:
      LOG(info) << ToString();
      LOG(info) << "\tDepends one three values : total hits in the following systems (in this order) : STS TRD TOF.";
      LOG(info) << "\tNegative values are interpreted as no-cut.";
      break;
    default:
      LOG(info) << ToString();
      LOG(info) << "\tNo user info.";
      break;
  }
}
//+++++++++++++++++++++  TrackFilter  ++++++++++++++++++++++++++++++++
//____________________________________________________________________
bool CbmRecoQaTask::TrackFilter::Accept(const CbmGlobalTrack* ptr, const CbmRecoQaTask* lnk)
{
  bool ret(true);
  stringstream ss;
  int idx(-1);
  const CbmTrack* trk(nullptr);
  switch (fType) {
    case eTrackCut::kSts:
      if (fNSts <= 0 && !fStsHits.size()) break;
      if ((idx = ptr->GetStsTrackIndex()) < 0) {
        ss << "Sts trk index missing";
        ret = false;
        break;
      }
      if (!(trk = lnk->GetTrack(ECbmModuleId::kSts, idx))) {
        ss << "Sts trk nullptr";
        ret = false;
        break;
      }
      // cut on the total no of STS hits/trk
      if (trk->GetNofHits() < fNSts) {
        ss << "Sts hits/trk [" << trk->GetNofHits() << "] < min[" << fNSts << "].";
        ret = false;
        break;
      }
      // cut on the distribution of STS hits/trk/station
      for (auto stsStation : fStsHits) {
        if (!stsStation) continue;
        // check hit in this particular station
      }
      break;
    case eTrackCut::kTrd:
      if (fNTrd <= 0) break;
      if ((idx = ptr->GetTrdTrackIndex()) < 0) {
        ss << "Trd trk index missing";
        ret = false;
        break;
      }
      if (!(trk = lnk->GetTrack(ECbmModuleId::kTrd, idx))) {
        ss << "Trd trk nullptr";
        ret = false;
        break;
      }
      // cut on the total no of TRD hits/trk
      if (trk->GetNofHits() < fNTrd) {
        ss << "Trd` hits/trk [" << trk->GetNofHits() << "] < min[" << fNTrd << "].";
        ret = false;
        break;
      }
      break;
    case eTrackCut::kTof:
      if (fNTof <= 0) break;
      if ((idx = ptr->GetTofTrackIndex()) < 0) {
        ss << "Tof trk index missing";
        ret = false;
        break;
      }
      if (!(trk = lnk->GetTrack(ECbmModuleId::kTof, idx))) {
        ss << "Tof trk nullptr";
        ret = false;
        break;
      }
      // cut on the total no of ToF hits/trk
      if (trk->GetNofHits() < fNTof) {
        ss << "Tof` hits/trk [" << trk->GetNofHits() << "] < min[" << fNTof << "].";
        ret = false;
        break;
      }
      break;
    default: break;
  }

  if (!ret) LOG(debug2) << "Track reject for : " << ss.str();
  return ret;
}
bool CbmRecoQaTask::TrackFilter::SetFilter(std::vector<float> cuts)
{
  switch (fType) {
    case eTrackCut::kSts:
      if (cuts.size() < 1) {
        LOG(warning) << "Improper definition for track filter :\n\t" << ToString() << endl;
        //HelpMess();
        return false;
      }
      fNSts = int(cuts[0]);
      break;
    case eTrackCut::kTrd:
      if (cuts.size() < 1) {
        LOG(warning) << "Improper definition for track filter :\n\t" << ToString() << endl;
        //HelpMess();
        return false;
      }
      fNTrd = int(cuts[0]);
      break;
    case eTrackCut::kTof:
      if (cuts.size() < 1) {
        LOG(warning) << "Improper definition for track filter :\n\t" << ToString() << endl;
        //HelpMess();
        return false;
      }
      fNTof = int(cuts[0]);
      break;
    default: break;
  }
  return true;
}
std::string CbmRecoQaTask::TrackFilter::ToString() const
{
  stringstream ss;
  switch (fType) {
    case eTrackCut::kSts: ss << "kSts : \"cut on no of STS hits / track\""; break;
    case eTrackCut::kMuch: ss << "kMuch : \"cut on no of Much hits / track\""; break;
    case eTrackCut::kRich: ss << "kRich : \"cut on no of Rich hits / track\""; break;
    case eTrackCut::kTrd: ss << "kTrd : \"cut on no of TRD hits / track\""; break;
    case eTrackCut::kTof: ss << "kTof : \"cut on no of Tof hits / track\""; break;
    default: break;
  }
  return ss.str();
}
/* clang-format off */
ClassImp(CbmRecoQaTask)
ClassImp(CbmRecoQaTask::Detector)
ClassImp(CbmRecoQaTask::Detector::View)
ClassImp(CbmRecoQaTask::TrackFilter)
ClassImp(CbmRecoQaTask::EventFilter)
  /* clang-format on */
